import abc


class SimulationPosteriorType(abc.ABC):
    @abc.abstractmethod
    def setup_posteriors(self):
        pass

    @abc.abstractmethod
    def inversion_mcmc(self):
        pass

    @abc.abstractmethod
    def inversion_optim_map(self):
        pass

    @abc.abstractmethod
    def inversion_optim_mle(self):
        pass



"""Implementation of Gaussian likelihood with censorship (with a lower limit)
"""

from typing import Optional, Union

import numpy as np
from scipy.stats import norm as statsnorm

from beetroots.modelling.likelihoods import utils
from beetroots.modelling.likelihoods.abstract_likelihood import Likelihood


class CensoredGaussianLikelihood(Likelihood):
    r"""Class implementing a Gaussian likelihood model with lower censorship"""

    __slots__ = (
        "forward_map",
        "D",
        "L",
        "N",
        "y",
        "sigma",
        "omega",
        "bias",
    )

    def __init__(
        self,
        forward_map,
        D: int,
        L: int,
        N: int,
        y: np.ndarray,
        sigma: Union[float, np.ndarray],
        omega: Union[float, np.ndarray],
        bias: Union[float, np.ndarray] = 0.0,
    ) -> None:
        """Constructor of the GaussianLikelihood object.

        Parameters
        ----------
        forward_map : ForwardMap instance
            forward map
        D : int
            number of disinct physical parameters in input space.
        L : int
            number of distinct observed physical parameters.
        N : int
            number of pixels in each physical dimension
        y : np.ndarray of shape (N, L)
            mean of the gaussian distribution
        bias : float or np.ndarray of shape (N, L)
            variance of the Gaussian distribution
        sigma : float or np.ndarray of shape (N, L)
            variance of the Gaussian distribution
        omega : float or np.ndarray of shape (N, L)
            censorship threshold

        Raises
        ------
        ValueError
            y must have the shape (N, L)
        """
        super().__init__(forward_map, D, L, N, y)

        # ! trigger an error is the mean y contains less than N elements
        if not (y.shape == (N, L)):
            raise ValueError(
                "y must have the shape (N, L) = ({}, {}) elements".format(
                    self.N, self.L
                )
            )
        if isinstance(sigma, (float, int)):
            self.sigma = sigma * np.ones((N, L))
        else:
            assert sigma.shape == (N, L)
            self.sigma = sigma

        if isinstance(omega, (float, int)):
            self.omega = omega * np.ones((N, L))
        else:
            assert omega.shape == (N, L)
            self.omega = omega

        if isinstance(bias, (float, int)):
            self.bias = bias * np.ones((N, L))
        else:
            assert bias.shape == (N, L)
            self.bias = bias

    def neglog_pdf(
        self,
        forward_map_evals: dict,
        nll_utils: dict,
        pixelwise: bool = False,
        full: bool = False,
        idx: Optional[np.ndarray] = None,
    ) -> Union[float, np.ndarray]:
        r"""[summary]

        .. math::
            p(y_{n,\ell} \vert \theta_n) \propto \exp \left\{- [y_{n,\ell} = \omega] \Phi( \frac{\omega - f_{\ell}(\theta_n)}{\sigma^2} \right) - [y_{n,\ell} > \omega] \frac{\omega - f_{\ell}(\theta_n)}{\sigma^2} \right\}
        """
        if idx is None:
            N_pix = self.N * 1
            y = self.y * 1
            sigma = self.sigma * 1
            omega = self.omega * 1
            bias = self.bias * 1

        else:
            n_pix = idx.size
            k_mtm = forward_map_evals["f_Theta"].shape[0] // n_pix
            N_pix = forward_map_evals["f_Theta"].shape[0]
            assert n_pix * k_mtm == N_pix

            y = np.zeros((n_pix, k_mtm, self.L))
            sigma = np.zeros((n_pix, k_mtm, self.L))
            omega = np.zeros((n_pix, k_mtm, self.L))
            bias = np.zeros((n_pix, k_mtm, self.L))

            for i_pix in range(n_pix):
                y[i_pix, :, :] = self.y[idx[i_pix], :][None, :] * np.ones(
                    (k_mtm, self.L)
                )
                sigma[i_pix, :, :] = self.sigma[idx[i_pix], :][None, :] * np.ones(
                    (k_mtm, self.L)
                )
                omega[i_pix, :, :] = self.omega[idx[i_pix], :][None, :] * np.ones(
                    (k_mtm, self.L)
                )
                bias[i_pix, :, :] = self.bias[idx[i_pix], :][None, :] * np.ones(
                    (k_mtm, self.L)
                )

            y = y.reshape((N_pix, self.L))
            sigma = sigma.reshape((N_pix, self.L))
            omega = omega.reshape((N_pix, self.L))
            bias = bias.reshape((N_pix, self.L))

        nlpdf = np.where(
            y <= omega,
            self.neglog_pdf_ac(forward_map_evals, nll_utils, y, sigma, omega, bias),
            self.neglog_pdf_au(forward_map_evals, nll_utils, y, sigma, omega, bias),
        )  # (N_pix, L)

        if full:
            return nlpdf  # (N_pix, L)

        if pixelwise:
            return np.sum(nlpdf, axis=1)  # (N_pix,)

        return np.sum(nlpdf)

    def neglog_pdf_ac(
        self,
        forward_map_evals: dict,
        nll_utils: dict,
        y: np.ndarray,
        sigma: np.ndarray,
        omega: np.ndarray,
        bias: np.ndarray,
    ) -> np.ndarray:
        return -statsnorm.logcdf((omega - forward_map_evals["f_Theta"] - bias) / sigma)

    def neglog_pdf_au(
        self,
        forward_map_evals: dict,
        nll_utils: dict,
        y: np.ndarray,
        sigma: np.ndarray,
        omega: np.ndarray,
        bias: np.ndarray,
    ) -> np.ndarray:
        return (forward_map_evals["f_Theta"] + bias - y) ** 2 / (2 * sigma**2)

    def gradient_neglog_pdf(
        self, forward_map_evals: dict, nll_utils: dict
    ) -> np.ndarray:
        """[summary]

        [extended_summary]

        Parameters
        ----------
        x : np.ndarray of shape (N, D)
            [description]
        f_Theta : np.ndarray of shape (N, L), optional
            image of x via forward map, by default None
        grad_f_Theta : np.ndarray of shape (N, D, L), optional
            [description], by default None

        Returns
        -------
        np.ndarray of shape (N, D)
            [description]
        """
        # if f_Theta is None:
        #     f_Theta = self.forward_map.evaluate(x)  # (N, L)
        # if grad_f_Theta is None:
        #     grad_f_Theta = self.forward_map.gradient(x)  # (N, D, L)

        grad_ = np.where(
            (self.y == self.omega)[:, None, :],
            self.gradient_neglog_pdf_ac(forward_map_evals, nll_utils),
            self.gradient_neglog_pdf_au(forward_map_evals, nll_utils),
        )  # (N, D, L)

        # ! issue: do not sum over L if L = D (i.e. identity forward_map)
        if not self.D == self.L:
            grad_ = np.sum(grad_, axis=2)  # (N, D)

        return grad_

    def gradient_neglog_pdf_ac(
        self, forward_map_evals: dict, nll_utils: dict
    ) -> np.ndarray:
        grad_ = (
            forward_map_evals["grad_f_Theta"]
            * (
                utils.norm_pdf_cdf_ratio(
                    (self.omega - forward_map_evals["f_Theta"] - self.bias) / self.sigma
                )
                / self.sigma
            )[:, None, :]
        )
        return grad_  # (N, D, L)

    def gradient_neglog_pdf_au(
        self, forward_map_evals: dict, nll_utils: dict
    ) -> np.ndarray:
        grad_ = (
            forward_map_evals["grad_f_Theta"]
            * ((forward_map_evals["f_Theta"] + self.bias - self.y) / self.sigma**2)[
                :, None, :
            ]
        )
        return grad_  # (N, D, L)

    def hess_diag_neglog_pdf(
        self, forward_map_evals: dict, nll_utils: dict
    ) -> np.ndarray:
        r"""[summary]

        [extended_summary]

        Parameters
        ----------
        x : np.ndarray of shape (N, D)
            [description]
        f_Theta : np.ndarray of shape (N, L), optional
            [description], by default None
        grad_f_Theta : np.ndarray of shape (N, D, L), optional
            [description], by default None
        hess_diag_f_Theta : np.ndarray of shape (N, D, L), optional
            [description], by default None

        Returns
        -------
        np.ndarray of shape (N, D, L)
            [description]
        """
        hess_diag = np.where(
            (self.y == self.omega)[:, None, :],
            self.hess_diag_neglog_pdf_ac(forward_map_evals, nll_utils),
            self.hess_diag_neglog_pdf_au(forward_map_evals, nll_utils),
        )  # (N, D, L)

        # ! issue: do not sum over L if L = D (i.e. identity forward_map)
        if not self.D == self.L:
            hess_diag = np.sum(hess_diag, axis=2)  # (N, D)

        return hess_diag

    def hess_diag_neglog_pdf_ac(
        self, forward_map_evals: dict, nll_utils: dict
    ) -> np.ndarray:
        hess_diag = (
            utils.norm_pdf_cdf_ratio(
                (self.omega - forward_map_evals["f_Theta"]) / self.sigma
            )
            / self.sigma
        )[:, None, :] * (
            forward_map_evals["hess_diag_f_Theta"]
            + forward_map_evals["grad_f_Theta"] ** 2
            * (
                (
                    (self.omega - forward_map_evals["f_Theta"] - self.bias) / self.sigma
                    + utils.norm_pdf_cdf_ratio(
                        (self.omega - forward_map_evals["f_Theta"] - self.bias)
                        / self.sigma
                    )
                )
                / self.sigma
            )[:, None, :]
        )
        return hess_diag  # (N, D, L)

    def hess_diag_neglog_pdf_au(
        self, forward_map_evals: dict, nll_utils: dict
    ) -> np.ndarray:
        return (1 / self.sigma**2)[:, None, :] * (
            forward_map_evals["grad_f_Theta"] ** 2
            + forward_map_evals["hess_diag_f_Theta"]
            * (forward_map_evals["f_Theta"] + self.bias - self.y)[:, None, :]
        )  # (N, D, L)

    def evaluate_all_nll_utils(
        self,
        forward_map_evals: dict,
        idx: Optional[np.ndarray] = None,
        compute_derivatives: bool = True,
        compute_derivatives_2nd_order: bool = True,
    ) -> dict:
        nll_utils = {}
        return nll_utils

    def sample_observation_model(
        self,
        forward_map_evals: dict,
        rng: np.random.Generator = np.random.default_rng(),
    ) -> np.ndarray:
        return np.maximum(
            self.omega,
            forward_map_evals["f_Theta"] + rng.normal(loc=0.0, scale=self.sigma),
        )

    def evaluate_all_forward_map(
        self,
        Theta: np.ndarray,
        compute_derivatives: bool,
        compute_derivatives_2nd_order: bool = True,
    ) -> dict:
        assert len(Theta.shape) == 2 and Theta.shape[1] == self.D
        forward_map_evals = self.forward_map.compute_all(
            Theta,
            compute_lin=True,
            compute_log=False,
            compute_derivatives=compute_derivatives,
            compute_derivatives_2nd_order=compute_derivatives_2nd_order,
        )
        return forward_map_evals



import os
from typing import List, Optional

import h5py
import numpy as np
import pandas as pd

from beetroots.inversion.plots.plots_estimator import PlotsEstimator
from beetroots.inversion.results.utils.abstract_util import ResultsUtil
from beetroots.inversion.results.utils.perf_saver import EstimatorPerfSaver
from beetroots.space_transform.abstract_transform import Scaler


class ResultsLowestObjective(ResultsUtil):

    __slots__ = (
        "model_name",
        "chain_type",
        "path_img",
        "path_data_csv_out",
        "N_run",
        "effective_T",
    )

    def __init__(
        self,
        model_name: str,
        chain_type: str,
        path_img: str,
        path_data_csv_out: str,
        N_run: int,
        T: int,
        freq_save: int,
    ):
        assert chain_type in ["mcmc", "optim_map", "optim_mle"]

        self.model_name = model_name
        self.chain_type = chain_type
        self.path_img = path_img
        self.path_data_csv_out = path_data_csv_out

        self.N_run = N_run
        self.effective_T = T // freq_save

    def read_data(
        self,
        list_chains_folders: List[str],
        idx_lowest_obj: int,
    ) -> np.ndarray:
        idx_chain = idx_lowest_obj // self.effective_T
        idx_inchain = idx_lowest_obj % self.effective_T

        assert 0 <= idx_chain < self.N_run
        assert 0 <= idx_inchain < self.effective_T

        for seed, mc_path in enumerate(list_chains_folders):
            if seed == idx_chain:
                with h5py.File(mc_path, "r") as f:
                    Theta_lowest_obj_lin = np.array(f["list_Theta"][idx_inchain])

        return Theta_lowest_obj_lin

    def create_folders(self) -> str:
        folder_path_init = f"{self.path_img}/estimators"
        folder_path_inter = f"{folder_path_init}/{self.model_name}"
        folder_path = f"{folder_path_inter}/MAP_{self.chain_type}"
        for path_ in [folder_path_init, folder_path_inter, folder_path]:
            if not os.path.isdir(path_):
                os.mkdir(path_)

        return folder_path

    def save_estimator_to_csv(
        self, Theta_lowest_obj_lin_full: np.ndarray, estimator_name: str
    ) -> None:
        # save estimator
        path_overall_results = f"{self.path_data_csv_out}/"
        path_overall_results += f"estimation_Theta_{self.model_name}_"
        path_overall_results += f"{estimator_name}.csv"

        N, D = Theta_lowest_obj_lin_full.shape

        df_MAP = pd.DataFrame()
        nn, dd = np.meshgrid(np.arange(N), np.arange(D))
        df_MAP["n"] = nn.astype(int).flatten()
        df_MAP["d"] = dd.astype(int).flatten()
        df_MAP = df_MAP.sort_values(by=["n", "d"])
        df_MAP[f"Theta_MAP_{self.chain_type}"] = Theta_lowest_obj_lin_full.flatten()
        df_MAP.to_csv(path_overall_results)
        return

    def main(
        self,
        list_chains_folders: List[str],
        lowest_obj: float,
        idx_lowest_obj: int,
        scaler: Scaler,
        Theta_true_scaled_full: Optional[np.ndarray],
        list_idx_sampling: List[int],
        list_fixed_values: np.ndarray,
        estimator_plot: Optional[PlotsEstimator],
    ) -> None:
        estimator_name = f"MAP_{self.chain_type}"

        Theta_lowest_obj_lin = self.read_data(
            list_chains_folders,
            idx_lowest_obj,
        )

        D = len(list_fixed_values)
        N, D_sampling = Theta_lowest_obj_lin.shape

        Theta_lowest_obj_lin_full = np.zeros((N, D))
        Theta_lowest_obj_lin_full[:, list_idx_sampling] += Theta_lowest_obj_lin

        for d in range(D):
            if list_fixed_values[d] is not None:
                Theta_lowest_obj_lin_full[:, d] += list_fixed_values[d]

        Theta_lowest_obj_scaled_full = scaler.from_lin_to_scaled(
            Theta_lowest_obj_lin_full,
        )

        # save the estimator itself
        self.save_estimator_to_csv(Theta_lowest_obj_lin_full, estimator_name)

        # save its performance
        perf_saver = EstimatorPerfSaver()
        if Theta_true_scaled_full is not None:
            mse = perf_saver.compute_MSE(
                Theta_lowest_obj_scaled_full,
                Theta_true_scaled_full,
            )
            snr = perf_saver.compute_SNR(
                Theta_lowest_obj_scaled_full,
                Theta_true_scaled_full,
            )
        else:
            mse = None
            snr = None

        perf_saver.save_estimator_performance(
            self.path_data_csv_out,
            estimator_name,
            self.model_name,
            mse,
            snr,
            lowest_obj,
        )

        # plot it
        N, _ = Theta_lowest_obj_lin.shape
        if N > 1 and estimator_plot is not None:
            folder_lowest_obj = self.create_folders()
            estimator_plot.plot_estimator(
                Theta_lowest_obj_lin_full,
                estimator_name,
                folder_lowest_obj,
                self.model_name,
            )
            # self.plot_map_nll_of_estimator(
            #     Theta_MAP_scaled, estimator_name, model_name
            # )
        return



from typing import List, Optional, Union

import numpy as np

from beetroots.modelling.forward_maps.abstract_exp import ExpForwardMap


class BasicExpForwardMap(ExpForwardMap):
    r"""Forward model such that for every pixel :math:`n \in [1, N]`

    .. math::
        f :  \theta_n \in \mathbb{R}^D \mapsto \exp(\theta_n) \in \mathbb{R}^D

    i.e. in this class, :math:`D = L`
    """

    def __init__(
        self, D, L, N, dict_fixed_values_scaled: dict[str, Optional[float]] = {}
    ):
        assert D == L
        super().__init__(D, L, N, dict_fixed_values_scaled)

        self.output_subset = np.arange(self.D)
        r"""List[int]: subset of outputs to be predicted. Can be updated with ``restrict_to_output_subset``"""

    def evaluate(self, Theta: np.ndarray) -> np.ndarray:
        assert Theta.shape[1] == self.D
        return np.exp(Theta)[:, self.output_subset]

    def evaluate_log(self, Theta: np.ndarray) -> np.ndarray:
        assert Theta.shape[1] == self.D
        return Theta[:, self.output_subset]

    def gradient(self, Theta: np.ndarray) -> np.ndarray:
        return np.exp(Theta)[:, None, self.output_subset] * np.ones(
            (self.N, self.D, self.L)
        )

    def gradient_log(self, Theta: np.ndarray) -> np.ndarray:
        return np.ones((self.N, self.D, self.L))

    def hess_diag(self, Theta: np.ndarray) -> np.ndarray:
        return np.exp(Theta)[:, None, self.output_subset] * np.ones(
            (self.N, self.D, self.L)
        )

    def hess_diag_log(self, Theta: np.ndarray) -> np.ndarray:
        return np.zeros((self.N, self.D, self.L))

    def compute_all(
        self,
        Theta: np.ndarray,
        compute_lin: bool = True,
        compute_log: bool = True,
        compute_derivatives: bool = True,
        compute_derivatives_2nd_order: bool = True,
    ) -> dict[str, np.ndarray]:
        forward_map_evals = dict()

        f_Theta = self.evaluate(Theta)[:, self.output_subset]

        #! not necessarily N (in candidates testing case for MTM)
        N_pix = f_Theta.shape[0]

        if compute_lin:
            forward_map_evals["f_Theta"] = f_Theta

        if compute_lin and compute_derivatives:
            grad_f_Theta = np.ones((N_pix, self.D, self.L)) * f_Theta[:, None, :]
            forward_map_evals["grad_f_Theta"] = grad_f_Theta

            if compute_derivatives_2nd_order:
                hess_diag_f_Theta = (
                    np.ones((N_pix, self.D, self.L)) * f_Theta[:, None, :]
                )
                forward_map_evals["hess_diag_f_Theta"] = hess_diag_f_Theta

        if compute_log:
            log_f_Theta = np.log(f_Theta)
            forward_map_evals["log_f_Theta"] = log_f_Theta

        if compute_log and compute_derivatives:
            grad_log_f_Theta = np.ones((N_pix, self.D, self.L))
            forward_map_evals["grad_log_f_Theta"] = grad_log_f_Theta

            if compute_derivatives_2nd_order:
                hess_diag_log_f_Theta = np.zeros((N_pix, self.D, self.L))
                forward_map_evals["hess_diag_log_f_Theta"] = hess_diag_log_f_Theta

        return forward_map_evals

    def restrict_to_output_subset(self, list_observables: List[int]) -> None:
        for idx in list_observables:
            assert 0 <= idx <= self.D

        self.output_subset = list_observables
        self.L = len(list_observables)



import copy
import multiprocessing as mp
import os
import time
from concurrent.futures import ProcessPoolExecutor
from typing import Optional

import numpy as np
import pandas as pd

from beetroots.inversion.run.abstract_run import Run
from beetroots.modelling.posterior import Posterior
from beetroots.sampler.abstract_sampler import Sampler
from beetroots.sampler.saver.abstract_saver import Saver
from beetroots.space_transform.abstract_transform import Scaler

# TODO : finish this class


class RunOptimMLE(Run):
    r"""class that runs inversions using an optimization approach, considering all pixels independently.

    .. warning::
        Unfinished class.
    """

    def __init__(self, path_data_csv_out: str, max_workers: int, N: int):
        r"""

        Parameters
        ----------
        path_data_csv_out : str
            path to the folder where results are to be saved
        max_workers : int
            max number of workers that can be used to run the inversion
        N : int
            number of pixels, i.e., of optimization problems to solve
        """
        self.path_data_csv_out = path_data_csv_out
        r"""str: path to the folder where results are to be saved"""

        self.max_workers = max_workers
        r"""int: max number of workers that can be used to run the inversion"""

        self.N = N
        r"""int: number of pixels, i.e., of optimization problems to solve"""

    def prepare_run(
        self,
        dict_posteriors: dict[str, Posterior],
        path_raw: str,
        N_runs: int,
    ) -> None:
        r"""prepares the run by creating empty folders to save the run results

        Parameters
        ----------
        dict_posteriors : dict[str, Posterior]
            dictionary of posterior distributions
        path_raw : str
            path to the folders where the ``.hdf5`` files are to be stored
        N_runs : int
            number of independent optimization runs to run per component :math:`n` / pixel
        """
        for model_name in list(dict_posteriors.keys()):
            for seed in range(N_runs):
                folder_path_inter = f"{path_raw}/{model_name}/"
                folder_path_inter += f"optim_MLE_{seed}"

                for n in range(self.N):
                    folder_path = f"{folder_path_inter}/pixel_{n}"
                    for path_ in [folder_path_inter, folder_path]:
                        if not os.path.isdir(path_):
                            os.mkdir(path_)

        return

    def run(
        self,
        dict_posteriors: dict[str, Posterior],
        sampler_: Sampler,
        saver_: Saver,
        N_runs: int,
        max_iter: int,
        path_raw: str,
        freq_save: int = 1,
    ) -> None:
        r"""runs the ``N`` inversions.

        .. warning::
            Unfinished method.

        Parameters
        ----------
        dict_posteriors : dict[str, Posterior]
            _dictionary of posterior distributions
        sampler_ : Sampler
            sampler to be used to generate the Markov chain(s)
        saver_ : Saver
            object responsible for progressively saving the Markov chain data during the run
        N_runs : int
            number of independent Markov chains to run per posterior distribution
        max_iter : int
            total duration of a Markov chain
        path_raw : str
            path to the folders where the ``.hdf5`` files are to be stored
        freq_save : int, optional
            frequency of saved iterates during the run (1 means that every iteration is saved), by default 1
        """
        # TODO: change this method using the old method run_optimization_MLE
        global _run_one_simulation_optim_mle_all_pixels

        def _run_one_simulation_optim_mle_all_pixels(dict_input: dict) -> dict:
            model_name = dict_input["model_name"]
            seed = dict_input["seed"]

            folder_path = f"{path_raw}/{model_name}/optim_MLE_{seed}"
            saver_seed = copy.deepcopy(saver_)
            saver_seed.set_results_path(folder_path)

            sampler_seed = copy.deepcopy(sampler_)

            tps0 = time.time()

            sampler_seed.sample(
                dict_posteriors[model_name],
                saver=saver_seed,
                max_iter=max_iter,
            )
            # return input dict with duration information
            dict_output = {
                "seed": seed,
                "model_name": model_name,
                "total_duration": time.time() - tps0,
            }
            return dict_output

        # * global function
        print("starting optimization MLE")
        list_params = [
            {"seed": seed, "model_name": model_name}
            for seed in range(N_runs)
            for model_name in list(dict_posteriors.keys())
        ]
        # TODO: multiprocessing and cuda?
        # with ProcessPoolExecutor(
        #     max_workers=self.max_workers, mp_context=mp.get_context("fork")
        # ) as p:
        #     list_simulations_durations = list(
        #         p.map(_run_one_simulation_optim_map_all_pixels, list_params)
        #     )

        # * non parallel version
        list_simulations_durations = []
        for params in list_params:
            duration = _run_one_simulation_optim_map_all_pixels(params)
            list_simulations_durations.append(duration)

        df_results_sampling = pd.DataFrame(list_simulations_durations)
        df_results_sampling.to_csv(f"{self.path_data_csv_out}/durations_optim_MLE.csv")
        print("optimization MLE done\n")
        return

    def main(
        self,
        dict_posteriors: dict[str, Posterior],
        sampler_: Sampler,
        saver_: Saver,
        N_runs: int,
        max_iter: int,
        path_raw: str,
        freq_save: int = 1,
    ) -> None:
        r"""sequentially calls ``prepare_run`` and ``run``

        Parameters
        ----------
        dict_posteriors : dict[str, Posterior]
            dictionary of posterior distributions
        sampler_ : Sampler
            optimizer
        saver_ : Saver
            object responsible for progressively saving the optimization run data during the run
        N_runs : int
            number of independent optimization runs to run per posterior distribution
        max_iter : int
            total duration of an optimization run
        path_raw : str
            path to the folders where the ``.hdf5`` files are to be stored
        freq_save : int, optional
            frequency of saved iterates during the run (1 means that every iteration is saved), by default 1
        """
        self.prepare_run(dict_posteriors, path_raw, N_runs)
        self.run(
            dict_posteriors,
            sampler_,
            saver_,
            N_runs,
            max_iter,
            path_raw,
            freq_save,
        )
        return

    #! Old method
    # def run_optimization_MLE(
    #     self,
    #     my_sampler_params,
    #     freq_save: int = 1,
    # ) -> None:

    #     global _run_one_simulation_one_pixel

    #     my_sampler_params.save_to_file(self.path_data_csv_in, "algo_params_optim_MLE.csv")

    #     def _run_one_simulation_one_pixel(dict_input: dict) -> dict:
    #         n = dict_input["n"]
    #         #! change seed between pixels
    #         #! interest: avoid starting at the same point for all n
    #         seed_folder = dict_input["seed"]
    #         seed_run = (dict_input["seed"] + 1) * (n + 1)
    #         model_name = dict_input["model_name"]

    #         idx_model = list(self.dict_posteriors.keys()).index(model_name)
    #         params = (self.list_mixing_model_params + self.list_gaussian_approx_params)[
    #             idx_model
    #         ]

    #         folder_path_inter = f"{path_raw}/{model_name}/opti_MLE_{seed_folder}"
    #         folder_path = f"{folder_path_inter}/pixel_{n}"

    #         saver_ = MySaver(
    #             1,
    #             self.D,
    #             self.L,
    #             folder_path,
    #             self.batch_size,
    #             freq_save=freq_save_mle,
    #         )
    #         sampler = MySampler(
    #             my_sampler_params,
    #             self.D,
    #             self.D_no_kappa,
    #             self.L,
    #             1,
    #             np.random.default_rng(seed_run),
    #         )
    #         y_pix = self.y[n, :].reshape((1, self.L))
    #         omega_pix = self.omega[n, :].reshape((1, self.L))

    #         if "mixing" in model_name:
    #             sigma_a_pix = self.sigma_a[n, :].reshape((1, self.L))
    #             sigma_m_pix = self.sigma_m[n, :].reshape((1, self.L))

    #             likelihood_1pix = MixingModelsLikelihood(
    #                 self.forward_map,
    #                 self.D,
    #                 self.L,
    #                 1,
    #                 y_pix,
    #                 sigma_a_pix,
    #                 sigma_m_pix,
    #                 omega_pix,
    #                 path_transition_params=params["path_transition_params"],
    #             )
    #         elif "gaussian_approx" in model_name:
    #             is_raw = params * 1
    #             if is_raw:
    #                 m_a_pix = 0
    #                 s_a_pix = self.sigma_a[n, :].reshape((1, self.L))
    #             else:
    #                 m_a = (np.exp(self.sigma_m ** 2 / 2) - 1) * self.y
    #                 s_a = np.sqrt(
    #                     self.y ** 2
    #                     * np.exp(self.sigma_m ** 2)
    #                     * (np.exp(self.sigma_m ** 2) - 1)
    #                     + self.sigma_a ** 2
    #                 )

    #                 m_a_pix = m_a[n, :].reshape((1, self.L))
    #                 s_a_pix = s_a[n, :].reshape((1, self.L))

    #             likelihood_1pix = CensoredGaussianLikelihood(
    #                 self.forward_map,
    #                 self.D,
    #                 self.L,
    #                 1,
    #                 y_pix,
    #                 s_a_pix,
    #                 omega_pix,
    #                 bias=m_a_pix,
    #             )
    #         else:
    #             raise ValueError(f"invalid model name : {model_name}")

    #         posterior_1pix = Posterior(
    #             self.D,
    #             self.L,
    #             1,
    #             likelihood=likelihood_1pix,
    #             prior_spatial=None,
    #             prior_indicator=self.prior_indicator_1pix,
    #         )

    #         tps0 = time.time()
    #         sampler.sample(
    #             posterior_1pix,
    #             saver=saver_,
    #             max_iter=self.T_OPTI_MLE,
    #             # sample_regu_weights=False,
    #             disable_progress_bar=True,
    #         )

    #         # return input dict with duration information
    #         total_duration = time.time() - tps0

    #         dict_output = {
    #             "n": n,
    #             "seed": seed_folder,
    #             "seed_run": seed_run,
    #             "model_name": model_name,
    #             "total_duration": total_duration,
    #         }
    #         return dict_output

    #     # * core function
    #     print("starting optimization MLE")
    #     list_params = [
    #         {"n": n, "seed": seed, "model_name": model_name}
    #         for n in range(self.N)
    #         for seed in range(N_runs)
    #         for model_name in list(self.dict_posteriors.keys())
    #     ]

    #     n_models = len(self.list_gaussian_approx_params)
    #     n_models += len(self.list_mixing_model_params)

    #     with ProcessPoolExecutor(
    #         max_workers=self.max_workers, mp_context=mp.get_context("fork")
    #     ) as p:
    #         list_simulations_durations = list(
    #             tqdm(
    #                 p.map(
    #                     _run_one_simulation_one_pixel,
    #                     list_params,
    #                 ),
    #                 total=self.N * N_runs * n_models,
    #             )
    #         )
    #     df_results_optim_mle = pd.DataFrame(list_simulations_durations)
    #     df_results_optim_mle.to_csv(
    #         f"{self.path_data_csv_out}/durations_optim_MLE.csv"
    #     )
    #     print("optimization MLE done")
    #     return



import os

import numpy as np

from beetroots.approx_optim.nn_bo import ApproxParamsOptimNNBO
from beetroots.simulations.astro.observation.abstract_real_data import (
    SimulationRealData,
)


class ReadDataRealData(SimulationRealData):
    def __init__(self, list_lines):
        """overwrite SimulationRealData __init__function"""
        self.list_lines_fit = list_lines


if __name__ == "__main__":
    path_data_cloud = f"{os.path.dirname(os.path.abspath(__file__))}"
    path_data_cloud += "/../../../data/orionbar"
    path_data_cloud = os.path.abspath(path_data_cloud)

    list_lines = [
        "co_v0_j11__v0_j10",  # 'CO_(11-10)',
        "co_v0_j12__v0_j11",  # 'CO_(12-11)',
        "co_v0_j13__v0_j12",  # 'CO_(13-12)',
        "co_v0_j14__v0_j13",  # 'CO_(14-13)',
        "co_v0_j15__v0_j14",  # 'CO_(15-14)',
        "co_v0_j16__v0_j15",  # 'CO_(16-15)',
        "co_v0_j17__v0_j16",  # 'CO_(17-16)',
        "co_v0_j18__v0_j17",  # 'CO_(18-17)',
        "co_v0_j19__v0_j18",  # 'CO_(19-18)',
        "co_v0_j20__v0_j19",  # 'CO_(20-19)',
        "co_v0_j21__v0_j20",  # 'CO_(21-20)',
        "co_v0_j23__v0_j22",  # 'CO_(23-22)',
        #
        # H2 lines
        "h2_v0_j2__v0_j0",  # 'H2_0-0_S(0)',
        "h2_v0_j3__v0_j1",  # 'H2_0-0_S(1)',
        "h2_v0_j4__v0_j2",  # 'H2_0-0_S(2)',
        "h2_v0_j5__v0_j3",  # 'H2_0-0_S(3)',
        "h2_v0_j6__v0_j4",  # 'H2_0-0_S(4)',
        "h2_v0_j7__v0_j5",  # 'H2_0-0_S(5)',
        #
        # CH+ lines
        "chp_j1__j0",  # 'CH+_(1-0)',
        "chp_j2__j1",  # 'CH+_(2-1)',
        "chp_j3__j2",  # 'CH+_(3-2)',
        "chp_j4__j3",  # 'CH+_(4-3)',
        "chp_j5__j4",  # 'CH+_(5-4)',
        "chp_j6__j5",  # 'CH+_(6-5)'
    ]

    # for orionbar, the additive noise variance changes from line to line
    (
        _,  # df_int_fit
        _,  # y_fit
        sigma_a,  # _fit
        _,  # omega_fit
        _,  # y_valid
        _,  # sigma_a_valid
        _,  # omega_valid
    ) = ReadDataRealData(list_lines).setup_observation(
        data_int_path=f"{path_data_cloud}/OrionBar_Joblin_Int.pkl",
        data_err_path=f"{path_data_cloud}/OrionBar_Joblin_Err.pkl",
        save_obs=False,
    )

    sigma_m = np.log(1.3)

    approx_optim = ApproxParamsOptimNNBO(
        list_lines,
        name="bo_nn_orionbar",
        D=4,
        D_no_kappa=3,
        K=30,
        log10_f_grid_size=100,
        N_samples_y=200_000,  # 250_000
        max_workers=20,
        sigma_a=sigma_a,
        sigma_m=sigma_m,
    )
    approx_optim.main(
        forward_model_name="meudon_pdr_model_dense",
        angle=60.0,
        lower_bounds_lin=np.array([1e-1, 1e5, 1e0, 1e0]),
        upper_bounds_lin=np.array([1e1, 1e9, 1e5, 4e1]),
        n_iter=40,
    )



"""Implementation of a full grid evaluation of the approximation parameters of the likelihood with both additive and multiplicative noises.
"""
import datetime
import json
import multiprocessing as mp
import os
import time
from concurrent.futures import ProcessPoolExecutor

import matplotlib.pyplot as plt
import numba as nb
import numpy as np
import pandas as pd
from scipy.integrate import simpson
from scipy.stats import gaussian_kde, ks_1samp
from tqdm.auto import tqdm

# import beetroots.utils as utils
from beetroots.modelling.forward_maps.regression_poly import PolynomialApprox
from beetroots.space_transform.transform import MyScaler

SMALLER_SIZE = 12
SMALL_SIZE = 16
MEDIUM_SIZE = 20
BIGGER_SIZE = 24

plt.rc("font", size=SMALL_SIZE)  # controls default text sizes
plt.rc("axes", titlesize=SMALL_SIZE)  # fontsize of the axes title
plt.rc("axes", labelsize=MEDIUM_SIZE)  # fontsize of the x and y labels
plt.rc("xtick", labelsize=SMALL_SIZE)  # fontsize of the tick labels
plt.rc("ytick", labelsize=SMALL_SIZE)  # fontsize of the tick labels
plt.rc("legend", fontsize=SMALLER_SIZE)  # legend fontsize
plt.rc("figure", titlesize=BIGGER_SIZE)  # fontsize of the figure title


P_lambda = np.poly1d(np.array([-6.0, 15.0, -10.0, 0.0, 0.0, 1.0]))
SQRT_2_PI = np.sqrt(2 * np.pi)
LOG_10 = np.log(10)


@nb.jit(nopython=True)
def neg_log_pdf_add(y, m_a, s_a):
    nll_au = 0.5 * ((y - m_a) / s_a) ** 2 + np.log(SQRT_2_PI * s_a)
    return nll_au


@nb.jit(nopython=True)
def neg_log_pdf_mult(y, m_m, s_m):
    nll_mu = np.where(
        y > 0,
        0.5 * ((np.log(y) - m_m) / s_m) ** 2 + np.log(y * SQRT_2_PI * s_m),
        np.infty,
    )
    return nll_mu


def compute_lambda(a0, a1, f_Theta_true):
    """carefull: a0 and a1 are in log, base 10, and computations require natural base

    Parameters
    ----------
    a0 : float
        center of mixing interval
    a1 : float
        radius of mixing interval
    f_Theta_true : float
        value of true f(x)

    Returns
    -------
    lambda: float
        mixing parameter
    """
    log_fm1 = (a0 - a1) * LOG_10
    log_fp1 = (a0 + a1) * LOG_10

    lambda_ = np.where(
        np.log(f_Theta_true) <= log_fm1,
        1,
        np.where(
            np.log(f_Theta_true) >= log_fp1,
            0,
            P_lambda((np.log(f_Theta_true) - log_fm1) / (log_fp1 - log_fm1)),
        ),
    )
    return lambda_


def pdf_my_approx(y, lambda_, m_a, s_a, m_m, s_m, f_Theta_true):
    neg_log_pdf_add_arr = neg_log_pdf_add(y, +m_a + f_Theta_true, s_a)
    neg_log_pdf_mult_arr = np.nan_to_num(
        neg_log_pdf_mult(y, +m_m + np.log(f_Theta_true), s_m)
    )

    log_res = neg_log_pdf_add_arr * lambda_ + neg_log_pdf_mult_arr * (1 - lambda_)
    return np.exp(-log_res)


def evaluate_pdf(y, a0, a1, f_Theta_true, sigma_a, sigma_m):
    lambda_ = compute_lambda(a0, a1, f_Theta_true)

    m_a = (np.exp(sigma_m**2 / 2) - 1) * f_Theta_true
    s_a = np.sqrt(
        f_Theta_true**2 * np.exp(sigma_m**2) * (np.exp(sigma_m**2) - 1)
        + sigma_a**2
    )

    combination = (sigma_a / f_Theta_true) * np.exp(-(sigma_m**2))
    m_m = -0.5 * np.log(1 + combination**2)
    s_m = np.sqrt(sigma_m**2 - 2 * m_m)

    my_pdf = pdf_my_approx(y, lambda_, m_a, s_a, m_m, s_m, f_Theta_true)
    return my_pdf


def evaluate_cdf(y, a0, a1, f_Theta_true, sigma_a, sigma_m):
    idx = np.argsort(y)
    y = y[idx]
    dy = np.diff(y, prepend=0)
    dy[0] = 0

    my_pdf = evaluate_pdf(y, a0, a1, f_Theta_true, sigma_a, sigma_m)
    my_pdf = my_pdf[idx]

    my_cdf = (my_pdf * dy).cumsum()
    my_cdf /= my_cdf[-1]  # normalize
    return my_cdf


def build_forward_model(
    D: int,
    D_no_kappa: int,
    deg: int,
    grid_name: str,
    ell: int,
):
    # load grid
    # grid_reg = utils.read_extraction_result_dat(
    #     f"{os.path.dirname(os.path.abspath(__file__))}/../../../data/grids/{grid_name}"
    # )
    # cols = list(grid_reg.columns)

    grid_reg["kappa"] = 1.0
    grid_reg = grid_reg[["kappa"] + cols]

    # define scaler and apply it
    scaler = MyScaler(X_grid_lin=grid_reg.iloc[:, :D].values, D_no_kappa=D_no_kappa)

    grid_reg.iloc[:, :D] = scaler.from_lin_to_scaled(grid_reg.iloc[:, :D].values)
    grid_reg = grid_reg.iloc[:, list(range(D)) + [D + ell]]

    L = len(grid_reg.columns) - D

    # fit forward model
    pdr_code = PolynomialApprox(grid_reg, D, D_no_kappa, L, deg)
    return pdr_code, grid_reg


def sample_theta(
    K: int,
    lower_bounds: np.ndarray,
    upper_bounds: np.ndarray,
    rng: np.random._generator.Generator,
) -> np.ndarray:
    r"""sample $\theta$ from Stratified MC in cube

    Parameters
    ----------
    K : int
        total number of samples per axis
    lower_bounds : np.ndarray of shape (D,)
        lower bounds of cube on $\theta$
    upper_bounds : np.ndarray of shape (D,)
        upper bounds of cube on $\theta$
    rng : np.random._generator.Generator
        random number generator

    Returns
    -------
    np.ndarray of shape (N_samples, D)
        $\theta$ samples
    """
    delta_upper_lower = np.array([upper_bounds[d] - lower_bounds[d] for d in range(D)])
    list_lower_bounds = [
        np.arange(K) / K * delta_upper_lower[d] + lower_bounds[d] for d in range(D)
    ]
    list_lower_bounds = np.meshgrid(*list_lower_bounds)
    list_lower_bounds = [
        lower_bounds_Theta_i.flatten() for lower_bounds_Theta_i in list_lower_bounds
    ]
    list_samples = []
    for d, lower_bounds_Theta_i in enumerate(list_lower_bounds):
        Vd = rng.uniform(
            low=lower_bounds_Theta_i,
            high=lower_bounds_Theta_i + delta_upper_lower[d] / K,
        )
        list_samples.append(Vd)
    x = np.vstack(list_samples).T
    return x


def plot_hist_log10_f_Theta(
    log10_f_Theta: np.ndarray, kde_log10_f_Theta, line: str, ell: int, output_path: str
) -> None:
    """plots histogram of log10(f(\theta))

    Parameters
    ----------
    log10_f_Theta : np.ndarray of shape (-1, 1)
        array of values of f(x) for considered line
    line : str
        name of the line
    ell : int
        index of the line
    output_path : str
        path of folder where the figure is to be saved
    """
    log10_f_Theta_low = log10_f_Theta.min()
    log10_f_Theta_high = log10_f_Theta.max()
    nbins = 200

    list_log10_f_Theta = np.linspace(log10_f_Theta_low, log10_f_Theta_high, nbins)

    pi_log10_f_Theta = kde_log10_f_Theta.pdf(list_log10_f_Theta)

    plt.figure(figsize=(8, 6))
    plt.title(f"line {ell} : {line}")
    plt.hist(
        log10_f_Theta,
        bins=nbins,
        range=(log10_f_Theta_low, log10_f_Theta_high),
        density=True,
        label="samples",
    )
    plt.plot(list_log10_f_Theta, pi_log10_f_Theta, "r-", label="KDE")
    plt.xlabel(r"$\log f_\ell (x)$")
    plt.grid()
    plt.legend()
    plt.tight_layout()
    plt.savefig(f"{output_path}/hist_log_f_Theta_ell{ell}.PNG")
    plt.close()


def estimate_avg_dks(dict_input: dict) -> dict:
    a0 = dict_input["a0"]
    a1 = dict_input["a1"]
    # log10_f_Theta = dict_input["log10_f_Theta"]
    list_log10_f_grid = dict_input["list_log10_f_grid"]
    sigma_a = dict_input["sigma_a"]
    sigma_m = dict_input["sigma_m"]
    N_samples_y = dict_input["N_samples_y"]
    pdf_kde_log10_f_Theta = dict_input["pdf_kde_log10_f_Theta"]

    list_f_grid = 10**list_log10_f_grid

    list_dks = np.zeros_like(list_log10_f_grid)
    for i, f_Theta_true in enumerate(list_f_grid):
        samples_a = np.random.normal(scale=sigma_a, size=N_samples_y)
        samples_m = np.random.lognormal(sigma=sigma_m, size=N_samples_y)

        samples_true = samples_m * f_Theta_true + samples_a
        samples_true.sort()

        statistic, _ = ks_1samp(
            samples_true,
            evaluate_cdf,
            args=(a0, a1, f_Theta_true, sigma_a, sigma_m),
        )
        list_dks[i] += statistic

    # Dks = interp1d(list_log10_f_grid, list_statistics, bounds_error=True)
    # list_dks = Dks(log10_f_Theta)

    #! average of Dks weighted with pdf of log10_f_Theta
    #! integrated between min and max of log10_f_Theta
    avg_dist = simpson(
        y=pdf_kde_log10_f_Theta * list_dks,
        x=list_log10_f_grid,
    )

    # list_dks = np.zeros_like(log10_f_Theta)
    # for ell in range(L):
    #     list_dks[:, ell] = Dks(log10_f_Theta[:, ell])

    dict_output = {
        "a0": a0,
        "a1": a1,
        "list_dists": list_dks,
        "avg_dist": avg_dist,
    }

    return dict_output


if __name__ == "__main__":
    tps0 = time.time()

    # * Problem settings
    # noise variances and considered line
    sigma_a = 1.38715e-10
    sigma_m = np.log(1.1)
    ell = 9

    # * forward model definition
    grid_name = "PDR17G1E20_P_cte_grid.dat"
    D = 4
    D_no_kappa = 3
    deg = 6

    forward_model, grid_reg = build_forward_model(D, D_no_kappa, deg, grid_name, ell)
    list_lines = list(grid_reg.columns[D:])
    L = len(list_lines) * 1
    assert L == 1
    line = list_lines[0]

    # * define bounds on \theta
    lower_bound_kappa = np.log(0.1)
    upper_bound_kappa = np.log(10)

    lower_bounds = np.array(
        [lower_bound_kappa] + list(grid_reg.iloc[:, 1:D].min().values)
    )
    upper_bounds = np.array(
        [upper_bound_kappa] + list(grid_reg.iloc[:, 1:D].max().values)
    )

    # * number of samples
    K = 30
    N_samples_theta = K**D
    r"""to build pdf of $P(\theta_\ell)$"""

    log10_f_grid_size = 100  # 250
    r"""number of points in grid on $P(\theta_\ell)$"""

    N_samples_y = 250_000  # 250_000
    r"""number of samples on $y_\ell$"""

    # * definition of grid of parameters to optimize
    # parameter to optimize : a = (center of interval, radius of interval)
    # (radius of interval = interval size /2)
    N_grid_a0 = 100  # 100
    N_grid_a1 = 50  # 50

    #! log10_f0 = log10(ratio std of noises), ie value of f at which noise
    #! variances are equal
    var_eps_m = np.exp(sigma_m**2) * (np.exp(sigma_m**2) - 1)
    log10_f0 = 0.5 * (2 * np.log(sigma_a) - np.log(var_eps_m)) / LOG_10

    bounds_a0_low = log10_f0 - 2
    bounds_a0_high = log10_f0 + 8

    bounds_a1_low = 0.01
    bounds_a1_high = 2

    list_a0 = np.linspace(bounds_a0_low, bounds_a0_high, N_grid_a0)
    list_a1 = np.linspace(bounds_a1_low, bounds_a1_high, N_grid_a1)

    # * create output folder
    now = datetime.datetime.now()
    dt_string = now.strftime("%Y-%m-%d_%H")

    output_path = os.path.dirname(
        os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
    )
    output_path = f"{output_path}/outputs/approx_optim_grid_{dt_string}"
    if not os.path.isdir(output_path):
        os.mkdir(output_path)

    # * save params of optimization
    optim_params = {
        "sigma_a": sigma_a,
        "sigma_m": sigma_m,
        "K": K,
        "N_samples_theta": N_samples_theta,
        "N_samples_y": N_samples_y,
        "log10_f_grid_size": log10_f_grid_size,
        "grid_a": {
            "a0": {"low": bounds_a0_low, "high": bounds_a0_high, "size": N_grid_a0},
            "a1": {"low": bounds_a1_low, "high": bounds_a1_high, "size": N_grid_a1},
        },
        "grid_name": grid_name,
        "ell": ell,
    }
    with open(f"{output_path}/optim_params.json", "w", encoding="utf-8") as f:
        json.dump(
            optim_params,
            f,
            ensure_ascii=False,
            indent=4,
        )
    print("params saved")

    # * sample X using Stratified MC and build distribution f(X)
    rng = np.random.default_rng(42)
    x = sample_theta(K, lower_bounds, upper_bounds, rng)

    # the division is to get log in base 10
    log10_f_Theta = forward_model.evaluate_log(x) / LOG_10
    assert log10_f_Theta.shape == (N_samples_theta, 1)
    log10_f_Theta = log10_f_Theta.flatten()

    log10_f_Theta_low = log10_f_Theta.min()
    log10_f_Theta_high = log10_f_Theta.max()

    list_log10_f_grid = np.linspace(
        log10_f_Theta_low, log10_f_Theta_high, log10_f_grid_size
    )

    kde_log10_f_Theta = gaussian_kde(log10_f_Theta)
    pdf_kde_log10_f_Theta = kde_log10_f_Theta.pdf(list_log10_f_grid)
    print("samples of x ready")

    plot_hist_log10_f_Theta(log10_f_Theta, kde_log10_f_Theta, line, ell, output_path)
    print("histograms of f(x) done")

    # * perform main computation: evaluate Kolmogorov distance on grid of f_Theta
    print("starting Kolmogorov Smirnov distances computations")
    list_params = [
        {
            "a0": a0,
            "a1": a1,
            "log10_f_Theta": log10_f_Theta,
            "list_log10_f_grid": list_log10_f_grid,
            "sigma_a": sigma_a,
            "sigma_m": sigma_m,
            "N_samples_y": N_samples_y,
            "pdf_kde_log10_f_Theta": pdf_kde_log10_f_Theta,
        }
        for a0 in list_a0
        for a1 in list_a1
    ]

    with ProcessPoolExecutor(max_workers=30, mp_context=mp.get_context("fork")) as p:
        list_results = list(
            tqdm(p.map(estimate_avg_dks, list_params), total=len(list_params))
        )
    print("Kolmogorov Smirnov distances computations done")

    # * save list_dists (in list_results) as table in csv
    list_dicts = []
    for dict_output in list_results:
        a0 = dict_output["a0"]
        a1 = dict_output["a1"]
        list_statistics = list(dict_output["list_dists"])
        dict_ = {"a0": a0, "a1": a1}
        for log10_f_Theta, statistic in zip(list_log10_f_grid, list_statistics):
            dict_[log10_f_Theta] = statistic

        list_dicts.append(dict_)

    df_statistics = pd.DataFrame(list_dicts)
    df_statistics.to_csv(f"{output_path}/results_statistics.csv")

    # * save values of avg dist
    list_dicts = []
    for dict_output in list_results:
        a0 = dict_output["a0"]
        a1 = dict_output["a1"]
        avg_dist = dict_output["avg_dist"]
        dict_ = {"a0": a0, "a1": a1, "avg_dist": avg_dist}
        list_dicts.append(dict_)

    df_avg_dist = pd.DataFrame(list_dicts)
    df_avg_dist.to_csv(f"{output_path}/results_avg_dist.csv")

    # * plot contourf of avg_dist in function of (a, alpha_f)
    AA0, AA1 = np.meshgrid(list_a0, list_a1)
    arr_avg_dist = np.zeros_like(AA0)

    for dict_output in list_results:
        a0 = dict_output["a0"]
        a1 = dict_output["a1"]
        avg_dist = dict_output["avg_dist"]

        idx_a0 = np.where(list_a0 == a0)[0][0]
        idx_a1 = np.where(list_a1 == a1)[0][0]
        # print(idx_a, N_a, idx_alpha_f, N_alpha_f)
        # print(arr_avg_dist.shape)
        arr_avg_dist[idx_a1, idx_a0] = avg_dist

    plt.figure(figsize=(8, 6))
    plt.title(f"neg log average KS distance on line {ell}")
    plt.contourf(AA0, AA1, -np.log10(arr_avg_dist), levels=200)
    plt.colorbar(format=lambda x, _: f"{x:.3f}")
    plt.xlabel(r"mixing interval center $(a_{\ell,1} + a_{\ell,0})/2$")
    plt.ylabel(r"mixing interval radius $(a_{\ell,1} - a_{\ell,0})/2$")
    plt.axvline(log10_f0, c="r", ls="--", label="equal variances")
    plt.legend()
    plt.tight_layout()
    plt.savefig(f"{output_path}/final_log.PNG")
    plt.close()

    plt.figure(figsize=(8, 6))
    plt.title("average KS distance on marginals")
    plt.contourf(AA0, AA1, arr_avg_dist, levels=200)
    plt.colorbar(format=lambda x, _: f"{x:1.2e}")
    plt.xlabel(r"mixing interval center $(a_{\ell,1} + a_{\ell,0})/2$")
    plt.ylabel(r"mixing interval radius $(a_{\ell,1} - a_{\ell,0})/2$")
    plt.axvline(log10_f0, c="r", ls="--", label="equal variances")
    plt.legend()
    plt.tight_layout()
    plt.savefig(f"{output_path}/final_lin.PNG")
    plt.close()

    duration = time.time() - tps0
    print(f"optimization of parameters finished. Total duration: {duration:.2f} s")



"""implementation of ESS computation, from :cite:t:`gelmanBayesianDataAnalysis2015a`

Original implementation : https://github.com/jwalton3141/jwalton3141.github.io/blob/master/assets/posts/ESS/rwmh.py

Corresponding post : https://jwalton.info/Efficient-effective-sample-size-python/
"""

import numba as nb
import numpy as np


@nb.jit(nopython=True)
def compute_ess(Theta: np.ndarray) -> float:
    r"""Computes the effective sample size of estimand of interest for one physical parameter :math:`\theta_{nd}`. Vectorized implementation.

    Parameters
    ----------
    Theta : np.ndarray of shape (m_chains, n_iters)
        values of the ``n_iters`` iterates for each of the ``m_chains`` Markov chains associated to one physical parameter :math:`\theta_{nd}`

    Returns
    -------
    float
        ESS for physical parameter :math:`\theta_{nd}`
    """

    # if len(Theta.shape) == 1:
    #     Theta = Theta.reshape((1, -1))
    # assert (
    #     len(Theta.shape) == 2
    # ), f"Theta has shape {Theta.shape}, should be a 2D array of shape (m_chains, n_iters)"

    m_chains, n_iters = Theta.shape

    variogram = lambda t: ((Theta[:, t:] - Theta[:, : (n_iters - t)]) ** 2).sum() / (
        m_chains * (n_iters - t)
    )

    post_var = gelman_rubin(Theta)

    t = 1
    rho = np.ones(n_iters)
    negative_autocorr = False

    # Iterate until the sum of consecutive estimates of autocorrelation is negative
    while not negative_autocorr and (t < n_iters):
        rho[t] = 1 - variogram(t) / (2 * post_var)

        if not t % 2:
            negative_autocorr = np.sum(rho[t - 1 : t + 1]) < 0

        t += 1

    return m_chains * n_iters / (1 + 2 * rho[1:t].sum())


@nb.jit(nopython=True)
def gelman_rubin(Theta: np.ndarray) -> float:
    r"""Estimates the marginal posterior variance. Vectorized implementation.

    Parameters
    ----------
    Theta : np.ndarray of shape (m_chains, n_iters)
        values of the ``n_iters`` iterates for each of the ``m_chains`` Markov chains associated to one physical parameter :math:`\theta_{nd}`

    Returns
    -------
    float
        marginal posterior variance
    """
    m_chains, n_iters = Theta.shape

    mean_arr = np.sum(Theta, axis=1) / n_iters
    if m_chains > 1:
        # Calculate between-chain variance
        B_over_n = ((mean_arr - np.mean(Theta)) ** 2).sum() / (m_chains - 1)
    else:
        B_over_n = 0.0

    # Calculate within-chain variances
    Theta_mean_same_dims = np.zeros_like(Theta)
    for chain in range(m_chains):
        Theta_mean_same_dims[chain, :] = mean_arr[chain]

    W = ((Theta - Theta_mean_same_dims) ** 2).sum() / (m_chains * (n_iters - 1))

    # (over) estimate of variance
    s2 = W * (n_iters - 1) / n_iters + B_over_n

    return s2



import os
from typing import List, Optional, Tuple

import h5py
import matplotlib.pyplot as plt
import numpy as np

from beetroots.inversion.results.utils.abstract_util import ResultsUtil


class ResultsObjective(ResultsUtil):

    __slots__ = (
        "model_name",
        "chain_type",
        "path_img",
        "N_MCMC",
        "effective_T_MC",
        "effective_T_BI",
    )

    def __init__(
        self,
        model_name: str,
        chain_type: str,
        path_img: str,
        N_MCMC: int,
        T_MC: int,
        T_BI: int,
        freq_save: int,
        N: int,
        D: int,
        L: int,
    ):
        assert chain_type in ["mcmc", "optim_map", "optim_mle"]

        self.model_name = model_name
        self.chain_type = chain_type
        self.path_img = path_img

        self.N_MCMC = N_MCMC
        self.effective_T_MC = T_MC // freq_save
        self.effective_T_BI = T_BI // freq_save

        self.N = N
        self.D = D
        self.L = L

    def read_data(
        self, list_chains_folders: List[str]
    ) -> Tuple[np.ndarray, np.ndarray]:
        list_objective = np.zeros((self.N_MCMC, self.effective_T_MC))
        list_nll = np.zeros((self.N_MCMC, self.effective_T_MC))
        # list_nl_prior_indicator = np.zeros((self.N_MCMC, self.effective_T_MC))
        # list_nl_prior_spatial = np.zeros((self.N_MCMC, self.effective_T_MC))

        for seed, mc_path in enumerate(list_chains_folders):
            with h5py.File(mc_path, "r") as f:
                list_objective[seed] = np.array(f["list_objective"])
                list_nll[seed] = np.array(f["list_nll"])
                # if "list_nl_prior_indicator" in f:
                #     list_nl_prior_indicator[seed] = np.array(
                #         f["list_nl_prior_indicator"]
                #     )
                # if "list_nl_prior_spatial" in f:
                #     list_nl_prior_spatial[seed] = np.array(f["list_nl_prior_spatial"])

        return list_objective, list_nll

    def create_folders(self) -> str:
        folder_path_inter = f"{self.path_img}/objective"
        folder_path = f"{folder_path_inter}/{self.model_name}"
        for path_ in [folder_path_inter, folder_path]:
            if not os.path.isdir(path_):
                os.mkdir(path_)
        return folder_path

    def plot_objective(
        self,
        folder_path: str,
        list_objective: np.ndarray,
        list_nll: np.ndarray,
        # list_nl_prior_indicator: np.ndarray,
        # list_nl_prior_spatial: np.ndarray,
        objective_true: Optional[float],
    ):
        filename_prefix = f"{folder_path}/{self.chain_type}_objective"

        list_objective_flat = list_objective.flatten()
        list_objective_no_BI_flat = list_objective[:, self.effective_T_BI :]
        list_objective_no_BI_flat = list_objective_no_BI_flat.flatten()

        list_nll_flat = list_nll.flatten()
        list_nll_no_BI_flat = list_nll[:, self.effective_T_BI :]
        list_nll_no_BI_flat = list_nll_no_BI_flat.flatten()

        # list_list_nl_prior_indicator_flat = []
        # list_list_nl_prior_indicator_no_BI_flat = []
        # for d in range(self.D):
        #     list_nl_prior_indicator_flat = list_nl_prior_indicator[:, :, d].flatten()

        #     list_nl_prior_indicator_no_BI_flat = list_nl_prior_indicator[
        #         :, self.effective_T_BI :, d
        #     ]
        #     list_nl_prior_indicator_no_BI_flat = (
        #         list_nl_prior_indicator_no_BI_flat.flatten()
        #     )

        # list_nl_prior_spatial_flat = list_nl_prior_spatial.flatten()
        # list_nl_prior_spatial_no_BI_flat = list_nl_prior_spatial[
        #     :, self.effective_T_BI :
        # ]
        # list_nl_prior_spatial_no_BI_flat = list_nl_prior_spatial_no_BI_flat.flatten()

        # * With Burn in
        plt.figure(figsize=(8, 6))
        plt.title("Objective evolution")
        plt.plot(list_objective_flat, label="objective")
        plt.plot(list_nll_flat, label="nll")

        for seed in range(self.N_MCMC):
            if seed == 0:
                plt.axvline(
                    seed * self.effective_T_MC + self.effective_T_BI,
                    c="k",
                    ls="--",
                    label="T_BI",
                )
            elif seed == 1:
                plt.axvline(
                    seed * self.effective_T_MC,
                    c="k",
                    ls="-",
                    label="new run",
                )
                plt.axvline(
                    seed * self.effective_T_MC + self.effective_T_BI,
                    c="k",
                    ls="--",
                )

            else:
                plt.axvline(seed * self.effective_T_MC, c="k", ls="-")
                plt.axvline(
                    seed * self.effective_T_MC + self.effective_T_BI,
                    c="k",
                    ls="--",
                )

        if list_objective.max() <= 0:
            plt.yscale("linear")
        elif list_objective.min() < 0:
            plt.yscale("symlog")
        else:
            plt.yscale("log")
        plt.grid()
        plt.legend()
        # plt.tight_layout()
        plt.savefig(
            f"{filename_prefix}_with_bi.PNG",
            bbox_inches="tight",
        )

        if objective_true is not None:
            plt.axhline(objective_true, c="r", ls="--", label="obj Theta_true")
            plt.legend()
            plt.savefig(
                f"{filename_prefix}_with_bi_and_true.PNG",
                bbox_inches="tight",
            )
        # if self.Theta_true_scaled is not None:
        #     forward_map_evals = self.dict_posteriors[
        #         model_name
        #     ].likelihood.evaluate_all_forward_map(self.Theta_true_scaled, True)
        #     nll_utils = self.dict_posteriors[
        #         model_name
        #     ].likelihood.evaluate_all_nll_utils(forward_map_evals)
        #     objective_true = self.dict_posteriors[model_name].neglog_pdf(
        #         self.Theta_true_scaled, forward_map_evals, nll_utils
        #     )
        #     plt.axhline(objective_true, c="r", ls="--", label="obj Theta_true")
        #     plt.legend()
        #     plt.savefig(
        #         f"{filename_prefix}_with_bi_and_true.PNG",
        #         bbox_inches="tight",
        #     )
        plt.close()

        # * Without Burn in
        plt.figure(figsize=(8, 6))
        plt.title("Objective evolution (no Burn-In)")
        plt.plot(list_objective_no_BI_flat, label="objective")
        plt.plot(list_nll_no_BI_flat, label="nll")

        for seed in range(1, self.N_MCMC):
            if seed == 1:
                plt.axvline(
                    seed * (self.effective_T_MC - self.effective_T_BI),
                    c="k",
                    ls="-",
                    label="new run",
                )
            else:
                plt.axvline(
                    seed * (self.effective_T_MC - self.effective_T_BI),
                    c="k",
                    ls="-",
                )

        if list_objective.max() < 0:
            plt.yscale("linear")
        elif list_objective.min() < 0:
            plt.yscale("symlog")
        else:
            plt.yscale("log")

        plt.grid()
        plt.legend()
        # plt.tight_layout()
        plt.savefig(
            f"{filename_prefix}_no_bi.PNG",
            bbox_inches="tight",
        )

        if objective_true is not None:
            plt.axhline(objective_true, c="r", ls="--", label="obj Theta_true")
            plt.legend()
            plt.savefig(
                f"{filename_prefix}_no_bi_with_true.PNG",
                bbox_inches="tight",
            )
        plt.close()
        return

    def plot_normalized_nll(
        self,
        folder_path: str,
        list_nll: np.ndarray,
    ) -> None:
        filename_prefix = f"{folder_path}/{self.chain_type}_nll"

        list_normalized_nll_flat = list_nll.flatten() / (self.N * self.L)
        list_normalized_nll_no_BI_flat = list_nll[:, self.effective_T_BI :] / (
            self.N * self.L
        )
        list_normalized_nll_no_BI_flat = list_normalized_nll_no_BI_flat.flatten()

        # * With Burn in
        plt.figure(figsize=(8, 6))
        plt.title("Normalized neglog likelihood evolution")
        plt.plot(list_normalized_nll_flat, label="nll")

        for seed in range(self.N_MCMC):
            if seed == 0:
                plt.axvline(
                    seed * self.effective_T_MC + self.effective_T_BI,
                    c="k",
                    ls="--",
                    label="T_BI",
                )
            elif seed == 1:
                plt.axvline(
                    seed * self.effective_T_MC,
                    c="k",
                    ls="-",
                    label="new run",
                )
                plt.axvline(
                    seed * self.effective_T_MC + self.effective_T_BI,
                    c="k",
                    ls="--",
                )

            else:
                plt.axvline(seed * self.effective_T_MC, c="k", ls="-")
                plt.axvline(
                    seed * self.effective_T_MC + self.effective_T_BI,
                    c="k",
                    ls="--",
                )

        if list_nll.max() <= 0:
            plt.yscale("linear")
        elif list_nll.min() < 0:
            plt.yscale("symlog")
        else:
            plt.yscale("log")
        plt.grid()
        plt.legend()
        # plt.tight_layout()
        plt.savefig(
            f"{filename_prefix}_with_bi.PNG",
            bbox_inches="tight",
        )
        plt.close()

        # * Without Burn in
        plt.figure(figsize=(8, 6))
        plt.title("Normalized neglog likelihood evolution (no Burn-In)")
        plt.plot(list_normalized_nll_no_BI_flat, label="nll")

        for seed in range(1, self.N_MCMC):
            if seed == 1:
                plt.axvline(
                    seed * (self.effective_T_MC - self.effective_T_BI),
                    c="k",
                    ls="-",
                    label="new run",
                )
            else:
                plt.axvline(
                    seed * (self.effective_T_MC - self.effective_T_BI),
                    c="k",
                    ls="-",
                )

        if list_nll.max() < 0:
            plt.yscale("linear")
        elif list_nll.min() < 0:
            plt.yscale("symlog")
        else:
            plt.yscale("log")

        plt.grid()
        plt.legend()
        # plt.tight_layout()
        plt.savefig(
            f"{filename_prefix}_no_bi.PNG",
            bbox_inches="tight",
        )
        plt.close()
        return

    def find_lowest_objective(
        self,
        list_objective: np.ndarray,
    ) -> Tuple[int, float]:
        # compute index of sampling MAP
        list_objective_no_BI = list_objective.flatten()
        idx_lowest_obj = int(np.argmin(list_objective_no_BI))
        lowest_obj = np.min(list_objective_no_BI)
        return idx_lowest_obj, lowest_obj

    def main(
        self,
        list_chains_folders: List[str],
        objective_true: Optional[float],
    ) -> Tuple[int, float]:
        list_objective, list_nll = self.read_data(list_chains_folders)
        folder_objective = self.create_folders()

        print("starting plot of objective function")
        self.plot_objective(
            folder_objective,
            list_objective,
            list_nll,
            objective_true,
        )
        self.plot_normalized_nll(folder_objective, list_nll)
        print("plot of objective function done")

        idx_lowest_obj, lowest_obj = self.find_lowest_objective(list_objective)
        return idx_lowest_obj, lowest_obj



from typing import List, Union

import numpy as np
from scipy.stats import gaussian_kde

from beetroots.approx_optim.abstract_approx_optim import ApproxParamsOptim
from beetroots.approx_optim.approach_type.bo import BayesianOptimizationApproach
from beetroots.approx_optim.forward_map.nn import ApproxOptimNN


class ApproxParamsOptimNNBO(
    ApproxParamsOptim, ApproxOptimNN, BayesianOptimizationApproach
):
    def main(
        self,
        dict_forward_model: str,
        lower_bounds_lin: Union[np.ndarray, List],
        upper_bounds_lin: Union[np.ndarray, List],
        n_iter: int,
    ):
        assert isinstance(forward_model_name, str)
        assert isinstance(angle, float)
        assert isinstance(lower_bounds_lin, (np.ndarray, list))
        assert isinstance(upper_bounds_lin, (np.ndarray, list))
        assert isinstance(n_iter, int) and n_iter > 0

        if isinstance(lower_bounds_lin, list):
            lower_bounds_lin = np.array(lower_bounds_lin)
        if isinstance(lower_bounds_lin, list):
            lower_bounds_lin = np.array(lower_bounds_lin)

        (
            log10_f0,
            bounds_a0_low,
            bounds_a0_high,
            bounds_a1_low,
            bounds_a1_high,
        ) = self.setup_params_bounds()

        log10_f_Theta = self.compute_log10_f_Theta(
            dict_forward_model,
            lower_bounds_lin,
            upper_bounds_lin,
        )
        log10_f_Theta_low = log10_f_Theta.min(axis=0)  # (L,)
        log10_f_Theta_high = log10_f_Theta.max(axis=0)  # (L,)

        list_log10_f_grid = np.zeros((self.log10_f_grid_size, self.L))
        pdf_kde_log10_f_Theta = np.zeros((self.log10_f_grid_size, self.L))
        for ell in range(self.L):
            list_log10_f_grid[:, ell] = np.linspace(
                log10_f_Theta_low[ell],
                log10_f_Theta_high[ell],
                self.log10_f_grid_size,
            )
            kde_log10_f_Theta = gaussian_kde(log10_f_Theta[:, ell])
            pdf_kde_log10_f_Theta[:, ell] = kde_log10_f_Theta.pdf(
                list_log10_f_grid[:, ell],
            )

        for ell in range(self.L):
            print(f"starting line {ell} ({self.list_lines[ell]})")
            self.plot_hist_log10_f_Theta(
                log10_f_Theta[:, ell],
                log10_f_Theta_low[ell],
                log10_f_Theta_high[ell],
                list_log10_f_grid[:, ell],
                pdf_kde_log10_f_Theta[:, ell],
                ell,
            )

            for n in range(self.N):
                pbounds = {
                    "a0": (bounds_a0_low[n, ell], bounds_a0_high[n, ell]),
                    "a1": (bounds_a1_low, bounds_a1_high),
                }
                self.save_setup_to_json(n, ell, pbounds)
                self.optimization(
                    first_points=[(log10_f0[n, ell], 0.5)],
                    init_points=3,
                    n_iter=n_iter,  # 50
                    list_log10_f_grid=list_log10_f_grid[:, ell],
                    pdf_kde_log10_f_Theta=pdf_kde_log10_f_Theta[:, ell],
                    pbounds=pbounds,
                    sigma_a_val=self.sigma_a[n, ell],
                    sigma_m_val=self.sigma_m[n, ell],
                    n=n,
                    ell=ell,
                )

        df_best = self.extract_optimal_params()
        df_best = df_best.set_index(["n", "ell"])

        for n in range(self.N):
            for ell in range(self.L):
                best_point = df_best.loc[(n, ell), ["a0_best", "a1_best"]].values
                self.plot_hist_log10_f_Theta_with_optim_results(
                    log10_f_Theta=log10_f_Theta[:, ell],
                    log10_f_Theta_low=log10_f_Theta_low[ell],
                    log10_f_Theta_high=log10_f_Theta_high[ell],
                    list_log10_f_grid=list_log10_f_grid[:, ell],
                    pdf_kde_log10_f_Theta=pdf_kde_log10_f_Theta[:, ell],
                    n=n,
                    ell=ell,
                    best_point=best_point,
                )

        self.plots_postprocessing(
            bounds_a0_low,
            bounds_a0_high,
            bounds_a1_low,
            bounds_a1_high,
            n_iter,
        )



import os
from typing import List

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from beetroots.inversion.results.utils.abstract_util import ResultsUtil


class ResultsValidMC(ResultsUtil):

    __slots__ = (
        "model_name",
        "path_img",
        "path_data_csv_out_mcmc",
        "N_MCMC",
        "effective_len_mc",
        "N",
        "D",
    )

    def __init__(
        self,
        model_name: str,
        path_img: str,
        path_data_csv_out_mcmc: str,
        N_MCMC: int,
        T_MC: int,
        T_BI: int,
        freq_save: int,
        N: int,
        D_sampling: int,
    ):
        self.model_name = model_name
        self.path_img = path_img
        self.path_data_csv_out_mcmc = path_data_csv_out_mcmc

        self.N_MCMC = N_MCMC
        self.effective_len_mc = (T_MC - T_BI) // freq_save

        self.N = N
        self.D_sampling = D_sampling

    def read_data(self) -> pd.DataFrame:
        path_file = f"{self.path_data_csv_out_mcmc}/"
        path_file += f"first_elt_valid_mc_{self.model_name}.csv"

        df_valid_mc = pd.read_csv(path_file)

        # assert len(df_valid_mc) == self.N_MCMC * self.N * self.D_sampling, f"len df_valid_mc={len(df_valid_mc)}, N_MCMC={self.N_MCMC}, N={self.N}, D_sampling={self.D_sampling}"
        return df_valid_mc

    def create_folders(self) -> str:
        folder_path_inter = f"{self.path_img}/well_reconstructed"
        folder_path = f"{folder_path_inter}/{self.model_name}"
        for path_ in [folder_path_inter, folder_path]:
            if not os.path.isdir(path_):
                os.mkdir(path_)

        return folder_path

    def main(
        self,
        list_names: List[str],
        list_idx_sampling: List[int],
    ) -> None:
        print("starting plot proportion of well reconstructed pixels")

        folder_path = self.create_folders()
        df_valid_mc = self.read_data()

        incr = 100 / self.N

        for seed in range(self.N_MCMC):
            df_seed = df_valid_mc[df_valid_mc["seed"] == seed]

            list_evolution = np.zeros((self.effective_len_mc, self.D_sampling))

            for idx_d, d in enumerate(list_idx_sampling):
                idx_arr_d = df_seed.loc[df_seed["d"] == d, "first_elt_valid_mc"].values
                for idx in idx_arr_d:
                    if idx > 0:
                        list_evolution[int(idx) :, idx_d] += incr

            plt.figure(figsize=(8, 6))
            plt.title("evolution of proportion of valid MC")
            idx_d = 0
            for d, name in enumerate(list_names):
                if d in list_idx_sampling:
                    plt.plot(list_evolution[:, idx_d], label=name)
                    idx_d += 1

            plt.plot(np.mean(list_evolution, 1), "k--", label="overall")
            plt.grid()
            plt.legend()

            filename = f"{folder_path}/prop_well_reconstruct_pixels_all_"
            filename += f"seed{seed}.PNG"
            plt.savefig(filename, bbox_inches="tight")
            plt.close()

        print("plot proportion of well reconstructed pixels done")
        return



r"""Contains a class that defines the transition between the sampling scale and user friendly / interpretable scale
"""
# from sklearn.preprocessing import StandardScaler
from typing import List, Optional

import numba
import numpy as np

from beetroots.space_transform.abstract_transform import Scaler


@numba.njit()
def _from_scaled_to_lin(
    Theta_scaled: np.ndarray,
    mean_: np.ndarray,
    std_: np.ndarray,
    D: int,
    D_no_kappa: int,
    LOG_10: float,
    list_is_log: List[bool],
) -> np.ndarray:
    # theta : rescale log and go back to linear scale
    Theta_linscale = np.zeros_like(Theta_scaled)

    # kappa : go back to linear scale
    if list_is_log[0]:
        Theta_linscale[:, 0] = np.exp(Theta_scaled[:, 0])
    else:
        Theta_linscale[:, 0] = Theta_scaled[:, 0] * 1

    # other params
    for d in range(1, D):
        rescaled = std_[d - 1] * Theta_scaled[:, d] + mean_[d - 1]
        if list_is_log[d]:
            Theta_linscale[:, d] = np.exp(LOG_10 * rescaled)
        else:
            Theta_linscale[:, d] = rescaled * 1

    return Theta_linscale


@numba.njit()
def _from_lin_to_scaled(
    Theta_linscale: np.ndarray,
    mean_: np.ndarray,
    std_: np.ndarray,
    D: int,
    D_no_kappa: int,
    list_is_log: List[bool],
) -> np.ndarray:
    # theta
    Theta_scaled = np.zeros_like(Theta_linscale)

    # kappa
    if list_is_log[0]:
        Theta_scaled[:, 0] = np.log(Theta_linscale[:, 0])
    else:
        Theta_scaled[:, 0] = Theta_linscale[:, 0] * 1

    # other params
    for d in range(1, D):
        if list_is_log[d]:
            scaled = np.log10(Theta_linscale[:, d])
        else:
            scaled = Theta_linscale[:, d] * 1

        Theta_scaled[:, d] = (scaled - mean_[d - 1]) / std_[d - 1]

    return Theta_scaled


class MyScaler(Scaler):
    r"""Defines the scale used during sampling and the transforms to navigate from one scale to the other"""

    __slots__ = ("D", "D_no_kappa", "mean_", "std_")
    LOG_10 = np.log(10.0)

    def __init__(
        self,
        Theta_grid_lin: Optional[np.ndarray] = None,
        D_no_kappa: Optional[int] = None,
        mean_=None,
        std_=None,
        list_is_log: Optional[List[bool]] = None,
    ):
        r"""

        Parameters
        ----------
        Theta_grid_lin : np.ndarray of shape (-1, D)
            grid of simulations
        D_no_kappa : int
            number of physical parameters that require a standard scaler
        """
        if mean_ is not None and std_ is not None and list_is_log is not None:
            self.D_no_kappa = mean_.size * 1
            r"""int: number of physical parameters that require a standard scaler"""
            self.D = self.D_no_kappa + 1
            r"""int: total number of physical parameters that require a standard scaler, including the scaling factor :math:`\kappa`"""
            self.mean_ = mean_
            r"""np.ndarray of shape (D,): mean of the D components :math:`\theta_d`, used in the data normalization"""
            self.std_ = std_
            r"""np.ndarray of shape (D,): standard deviation of the D components :math:`\theta_d`, used in the data normalization"""

            assert len(list_is_log) == self.D, f"{self.D}, {len(list_is_log)}"
            self.list_is_log = list_is_log

        else:
            self.D = Theta_grid_lin.shape[1]
            self.D_no_kappa = D_no_kappa if D_no_kappa is not None else self.D
            assert self.D_no_kappa <= self.D

            assert isinstance(list_is_log, list) and len(list_is_log) == self.D
            self.list_is_log = list_is_log

            raise NotImplementedError()

            # TODO : correct this condition
            log10_grid_theta = np.log10(Theta_grid_lin[:, (self.D - self.D_no_kappa) :])
            self.mean_ = log10_grid_theta.mean(axis=0)  # (D,)
            self.std_ = log10_grid_theta.std(axis=0)  # (D,)

        assert self.mean_.shape == (
            self.D_no_kappa,
        ), f"{self.D_no_kappa}, {self.mean_.shape}"
        assert self.std_.shape == (
            self.D_no_kappa,
        ), f"{self.D_no_kappa}, {self.std_.shape}"

    def from_scaled_to_lin(self, Theta_scaled: np.ndarray) -> np.ndarray:
        assert len(Theta_scaled.shape) == 2, Theta_scaled.shape
        assert Theta_scaled.shape[1] == self.D, Theta_scaled.shape

        Theta_linscale = _from_scaled_to_lin(
            Theta_scaled,
            self.mean_,
            self.std_,
            self.D,
            self.D_no_kappa,
            self.LOG_10,
            self.list_is_log,
        )
        return Theta_linscale

    def from_lin_to_scaled(self, Theta_linscale: np.ndarray) -> np.ndarray:
        assert len(Theta_linscale.shape) == 2, Theta_linscale.shape
        assert Theta_linscale.shape[1] == self.D, Theta_linscale.shape
        Theta_scaled = _from_lin_to_scaled(
            Theta_linscale,
            self.mean_,
            self.std_,
            self.D,
            self.D_no_kappa,
            self.list_is_log,
        )
        return Theta_scaled



# from typing import Callable

import numpy as np

from beetroots.approx_optim.approach_type.abstract_approach_type import ApproachType


class GridSearchApproach(ApproachType):
    def optimization(
        self,
        first_points: list[np.ndarray],
        init_points: int,
        n_iter: int,
    ):
        pass

    def plot_cost_map(self):
        pass



"""Abstract forward map
"""
from abc import ABC, abstractmethod
from typing import List, Optional, Union

import numpy as np


class ForwardMap(ABC):
    r"""Forward model such that for every pixel :math:`n \in [1, N]`

    .. math::

        f : \theta_n \in \mathbb{R}^D \mapsto f(\theta_n) \in \mathbb{R}^L
    """

    __slots__ = (
        "D",
        "L",
        "N",
        "dict_fixed_values_scaled",
        "list_indices_to_sample",
        "list_indices_fixed",
        "D_sampling",
        "arr_fixed_values",
    )

    def __init__(
        self,
        D: int,
        L: int,
        N: int,
        dict_fixed_values_scaled: dict[str, Optional[float]] = {},
    ) -> None:
        self.D = D
        """int: dimensionality of input space"""
        self.L = L
        """int: dimensionality of output space"""
        self.N = N
        """int: number of independent pixels"""

        self.set_sampled_and_fixed_entries(dict_fixed_values_scaled)

    def set_sampled_and_fixed_entries(
        self, dict_fixed_values_scaled: dict[str, Optional[float]] = {}
    ) -> None:
        # manually fixed values, constant during sampling / optim
        self.dict_fixed_values_scaled = dict_fixed_values_scaled
        r"""dict: indices of the entries to be fixed and the associated value"""

        self.list_indices_to_sample = [
            d
            for d, value in enumerate(dict_fixed_values_scaled.values())
            if value is None
        ]
        r"""list: indices of the entries to be sampled"""

        self.list_indices_fixed = [
            d
            for d, value in enumerate(dict_fixed_values_scaled.values())
            if value is not None
        ]
        r"""list: indices of the entries to be fixed"""

        self.D_sampling = len(self.list_indices_to_sample)
        r"""int: dimension of the subspace to sample (considering fixed values)"""

        arr_fixed_values = np.zeros((self.D,))
        for d, value in enumerate(dict_fixed_values_scaled.values()):
            if value is not None:
                arr_fixed_values[d] = value * 1
        self.arr_fixed_values = arr_fixed_values
        r"""np.ndarray: values of :math:`\theta` that are not sampled, but set to a specific value. The indices of fixed entries are given in ``list_indices_fixed``"""

    @abstractmethod
    def evaluate(self, Theta: np.ndarray) -> np.ndarray:
        r"""evaluates the forward map for a set of input vectors :math:`(\theta_n))_{n=1}^N`

        Parameters
        ----------
        Theta : np.ndarray of shape (N, D)
            array of points in the input space :math:`\Theta = (\theta_n)_{n=1}^N` with :math:`\theta_n \in \mathbb{R}^D`

        Returns
        -------
        np.ndarray of shape (N, L)
            array of the forward map images in the observation space :math:`(f(\theta_n))_{n=1}^N` with :math:`f(\theta_n) \in \mathbb{R}^L`
        """
        pass

    @abstractmethod
    def gradient(self, Theta: np.ndarray) -> np.ndarray:
        r"""returns the gradient of the forward map for a set of input vectors :math:`(\theta_n))_{n=1}^N`

        Parameters
        ----------
        Theta : np.ndarray of shape (N, D)
            array of points in the input space :math:`\Theta = (\theta_n)_{n=1}^N` with :math:`\theta_n \in \mathbb{R}^D`

        Returns
        -------
        np.ndarray of shape (N, D, L)
            array of gradients :math:`(\nabla f(\theta_n))_{n=1}^N`, with :math:`\nabla f(\theta_n) \in \mathbb{R}^{D \times L}`
        """
        pass

    @abstractmethod
    def hess_diag(self, Theta: np.ndarray) -> np.ndarray:
        r"""returns the diagonal of the hessian of the forward map for a set of input vectors :math:`(\theta_n))_{n=1}^N`

        Parameters
        ----------
        Theta : np.ndarray of shape (N, D)
            array of points in the input space :math:`\Theta = (\theta_n)_{n=1}^N` with :math:`\theta_n \in \mathbb{R}^D`

        Returns
        -------
        np.ndarray of shape (N, D, L)
            array of diagonal of the hessians :math:`(\text{diag} \nabla^2 f(\theta_n))_{n=1}^N`, with :math:`\nabla f(\theta_n) \in \mathbb{R}^{D \times L}`
        """
        pass

    @abstractmethod
    def compute_all(
        self,
        Theta: np.ndarray,
        compute_lin: bool = True,
        compute_log: bool = True,
        compute_derivatives: bool = True,
        compute_derivatives_2nd_order: bool = True,
    ) -> dict[str, np.ndarray]:
        r"""gathers the evaluation of the forward map in linear and log scale and of the associated derivatives. Permits to limit repeating computations, but requires the storage in memory of the result.

        Parameters
        ----------
        Theta : np.ndarray of shape (N, D)
            array of points in the input space :math:`\Theta = (\theta_n)_{n=1}^N` with :math:`\theta_n \in \mathbb{R}^D`
        compute_lin : bool, optional
            wether or not to compute the forward model (and possibly the gradient and diagonal of the Hessian), by default True
        compute_log : bool, optional
            wether or not to compute the log-forward model (and possibly the gradient and diagonal of the Hessian), by default True
        compute_derivatives : bool, optional
            wether or not to evaluate the derivatives of the forward map, by default True
        compute_derivatives_2nd_order : bool, optional
            wether or not to evaluate the 2nd order derivatives of the forward map, by default True

        Returns
        -------
        forward_map_evals : dict[str, np.ndarray]
            dictionary with entries such as `f_Theta`, `log_f_Theta`, `grad_f_Theta`, `grad_log_f_Theta`, `hess_diag_f_Theta` and `hess_diag_log_f_Theta`, depending on the input booleans.
        """
        pass

    @abstractmethod
    def restrict_to_output_subset(self, list_observables: List[Union[int, str]]):
        r"""restricts the list of outputs to be predicted to a subset, either identified by their indices or by names

        Parameters
        ----------
        list_observables : List[Union[int, str]]
            subset of outputs to be predicted
        """
        pass



"""Utils functions for censored likelihoods
"""
from typing import Union, overload

import numpy as np
from scipy.special import log_ndtr


@overload
def logpdf_normal(x: np.ndarray) -> np.ndarray:
    ...


@overload
def logpdf_normal(x: Union[float, int]) -> float:
    ...


def logpdf_normal(Theta: Union[np.ndarray, float, int]) -> Union[np.ndarray, float]:
    """log pdf of the standard gaussian distribution

    Parameters
    ----------
    Theta : np.ndarray
        points at which the function is to be evaluated in a vectorized way

    Returns
    -------
    np.ndarray
        log pdf of the standard gaussian distribution
    """
    return -(Theta**2) / 2 - 0.5 * np.log(2 * np.pi)


@overload
def norm_pdf_cdf_ratio(Theta: np.ndarray) -> np.ndarray:
    ...


@overload
def norm_pdf_cdf_ratio(Theta: Union[float, int]) -> float:
    ...


def norm_pdf_cdf_ratio(
    Theta: Union[np.ndarray, float, int]
) -> Union[np.ndarray, float]:
    r"""computes the ratio of the pdf and cdf of the standard gaussian distribution at a given point

    Parameters
    ----------
    Theta : float or np.array
        current point

    Returns
    -------
    float or np.array
        ratio of the pdf and cdf of the standard gaussian distribution (has the same shape as Theta, if Theta is a np.array)
    """
    return np.exp(logpdf_normal(Theta) - log_ndtr(Theta))



from typing import List, Union

import numpy as np


class SpatialPriorParams:
    r"""Dataclass to contain data on the spatial prior"""

    __slots__ = (
        "name",
        "use_next_nearest_neighbors",
        "initial_regu_weights",
    )

    def __init__(
        self,
        name: str,
        use_next_nearest_neighbors: bool,
        initial_regu_weights: Union[np.ndarray, List[float]],
    ) -> None:
        r"""

        Parameters
        ----------
        name : str
            name of the spatial regularization type, must be an element of ["L2-laplacian", "L2-gradient"]
        use_next_nearest_neighbors : bool
            wether or not to use the next nearest neighbors, i.e., in diagonal
        initial_regu_weights : Union[np.ndarray, List[float]]
            initial regularization weights (the regularization weights can be tuned automatically during the Markov chain)
        """
        assert name in ["L2-laplacian", "L2-gradient"]

        self.name = name
        r"""str: name of the spatial regularization type, must be an element of ["L2-laplacian", "L2-gradient"]"""

        self.use_next_nearest_neighbors = use_next_nearest_neighbors
        r"""bool: wether or not to use the next nearest neighbors, i.e., in diagonal"""

        self.initial_regu_weights = initial_regu_weights
        r"""Union[np.ndarray, List[float]: initial regularization weights (the regularization weights can be tuned automatically during the Markov chain)"""



"""Defines the Total variation spatial regularization and its derivatives (for 1D signals only)
"""
import numpy as np
import pandas as pd

from beetroots.modelling.priors.abstract_spatial_prior import SpatialPrior


class TVeps1DSpatialPrior(SpatialPrior):
    def __init__(
        self,
        D: int,
        N: int,
        df: pd.DataFrame,
        weights: np.ndarray = None,
        eps: float = 1e-3,
    ):
        super().__init__(D, N, df, weights)
        self.eps = eps

    def neglog_pdf(self, x: np.ndarray, with_weights: bool = True) -> np.ndarray:
        assert x.shape == (self.N, self.D)

        if self.list_edges.size == 0:
            return np.zeros((self.D,))

        nlpdf = np.sum(
            np.sqrt(
                (x[self.list_edges[:, 1]] - x[self.list_edges[:, 0]]) ** 2 + self.eps
            ),
            axis=0,
        )
        if with_weights:
            nlpdf *= self.weights

        return nlpdf  # (D,)

    def gradient_neglog_pdf(self, x: np.ndarray) -> np.ndarray:
        assert x.shape == (self.N, self.D)
        grad_ = np.zeros_like(x, dtype=np.float64)
        if self.list_edges.size == 0:
            return grad_

        for edge in self.list_edges:
            val = (x[edge[1]] - x[edge[0]]) / np.sqrt(
                (x[edge[1]] - x[edge[0]]) ** 2 + self.eps
            )  # (D,)
            grad_[edge[0], :] -= val
            grad_[edge[1], :] += val

        grad_ = grad_ * self.weights[None, :]
        return grad_  # (N, D)

    def hess_diag_neglog_pdf(self, x: np.ndarray) -> np.ndarray:
        assert x.shape == (self.N, self.D)

        hess_diag = np.zeros_like(x, dtype=np.float64)

        for edge in self.list_edges:
            # val = 1 / np.sqrt((x[edge[1]] - x[edge[0]]) ** 2 + self.eps) - (
            #     x[edge[1]] - x[edge[0]]
            # ) ** 2 * ((x[edge[1]] - x[edge[0]]) ** 2 + self.eps) ** (
            #     -3 / 2
            # )  # (D,)
            delta_x = x[edge[1]] - x[edge[0]]
            val = 2 * self.eps / (delta_x**2 + self.eps) ** (3 / 2)
            hess_diag[edge[0], :] += val
            hess_diag[edge[1], :] += val

        hess_diag = hess_diag * self.weights[None, :]
        return hess_diag



import numpy as np

from beetroots.approx_optim.forward_map.abstract_forward_map import (
    ApproxOptimForwardMap,
)
from beetroots.simulations.astro.forward_map.abstract_poly_reg import (
    SimulationPolynomialReg,
)


class ApproxOptimPolynomialReg(ApproxOptimForwardMap):
    def compute_log10_f_Theta(
        self,
        forward_model_name: str,
        angle: float,
        lower_bounds_lin: np.ndarray,
        upper_bounds_lin: np.ndarray,
    ):
        simulation = SimulationPolynomialReg()
        simulation.list_lines_fit = self.list_lines * 1
        scaler, forward_map = simulation.setup_forward_map(
            forward_model_name=forward_model_name,
            force_use_cpu=False,
            angle=angle,
        )

        lower_bounds = scaler.from_lin_to_scaled(
            lower_bounds_lin.reshape((1, self.D)),
        ).flatten()
        upper_bounds = scaler.from_lin_to_scaled(
            upper_bounds_lin.reshape((1, self.D)),
        ).flatten()
        x = self.sample_theta(lower_bounds, upper_bounds)

        # the division is to get log in base 10
        log10_f_Theta = forward_map.evaluate_log(x) / np.log(10)
        assert log10_f_Theta.shape == (self.N_samples_theta, self.L)
        return log10_f_Theta



"""Implementation of the identity forward map
"""
from typing import List, Optional

import numpy as np

from beetroots.modelling.forward_maps.abstract_base import ForwardMap


class BasicForwardMap(ForwardMap):
    r"""Forward model such that for every pixel :math:`n \in [1, N]`

    .. math::
        f :  \theta_n \in \mathbb{R}^D \mapsto \theta_n \in \mathbb{R}^D

    i.e. in this class, :math:`D = L`
    """

    def __init__(
        self, L: int, N: int, dict_fixed_values_scaled: dict[str, Optional[float]] = {}
    ):
        super().__init__(L, L, N, dict_fixed_values_scaled)

        self.output_subset = np.arange(self.D)
        r"""List[int]: subset of outputs to be predicted. Can be updated with ``restrict_to_output_subset``"""

    def evaluate(self, Theta: np.ndarray) -> np.ndarray:
        return Theta[:, self.output_subset]  # (N, L)

    def gradient(self, Theta: np.ndarray) -> np.ndarray:
        return np.ones((Theta.shape[0], self.L, self.L))  # (N, D, L)

    def hess_diag(self, Theta: np.ndarray) -> np.ndarray:
        return np.zeros((Theta.shape[0], self.L, self.L))

    def compute_all(
        self,
        Theta: np.ndarray,
        compute_lin: bool = True,
        compute_log: bool = False,
        compute_derivatives: bool = True,
        compute_derivatives_2nd_order: bool = True,
    ) -> dict:
        r"""gathers the evaluation of the forward map in linear and log scale and of the associated derivatives. Permits to limit repeating computations, but requires the storage in memory of the result.

        Parameters
        ----------
        Theta : np.ndarray of shape (N, D)
            array of points in the input space :math:`\Theta = (\theta_n)_{n=1}^N` with :math:`\theta_n \in \mathbb{R}^D`
        compute_lin : bool, optional
            always considered as True. Kept for this class for consistency.
        compute_log : bool, optional
            always considered as False. Kept for this class for consistency.
        compute_derivatives : bool, optional
            wether or not to evaluate the derivatives of the forward map, by default True
        compute_derivatives_2nd_order : bool, optional
            wether or not to evaluate the 2nd order derivatives of the forward map, by default True

        Returns
        -------
        forward_map_evals : dict[str, np.ndarray]
            dictionary with the `f_Theta` entry and possibly `grad_f_Theta`, and `hess_diag_f_Theta`, depending on the input booleans.
        """
        forward_map_evals = dict()
        forward_map_evals["f_Theta"] = self.evaluate(Theta)[:, self.output_subset]

        if compute_derivatives:
            forward_map_evals["grad_f_Theta"] = self.gradient(Theta)

            if compute_derivatives_2nd_order:
                forward_map_evals["hess_diag_f_Theta"] = self.hess_diag(Theta)

        return forward_map_evals

    def restrict_to_output_subset(self, list_observables: List[int]) -> None:
        for idx in list_observables:
            assert 0 <= idx <= self.D

        self.output_subset = list_observables
        self.L = len(list_observables)



import numpy as np

from beetroots.approx_optim.nn_bo import ApproxParamsOptimNNBO

if __name__ == "__main__":
    # for toycase, the noise variances are constant all over the map
    # and for all lines
    sigma_a = 1.38715e-10
    sigma_m = np.log(1.1)

    list_lines = [
        "co_v0_j4__v0_j3",
        "co_v0_j5__v0_j4",
        "co_v0_j6__v0_j5",
        "co_v0_j7__v0_j6",
        "co_v0_j8__v0_j7",
        "co_v0_j9__v0_j8",
        "co_v0_j10__v0_j9",
        "co_v0_j11__v0_j10",
        "co_v0_j12__v0_j11",
        "co_v0_j13__v0_j12",
    ]

    approx_optim = ApproxParamsOptimNNBO(
        list_lines,
        name="bo_nn_toycase",
        D=4,
        D_no_kappa=3,
        K=30,
        log10_f_grid_size=100,
        N_samples_y=200_000,  # 250_000
        max_workers=20,
        sigma_a=sigma_a,
        sigma_m=sigma_m,
    )
    approx_optim.main(
        forward_model_name="meudon_pdr_model_dense",
        angle=0.0,
        lower_bounds_lin=np.array([1e-1, 1e5, 1e0, 1e0]),
        upper_bounds_lin=np.array([1e1, 1e9, 1e5, 4e1]),
        n_iter=40,
    )



import os
import sys

import numpy as np

from beetroots.approx_optim.nn_bo import ApproxParamsOptimNNBO
from beetroots.simulations.astro.observation.abstract_real_data import (
    SimulationRealData,
)


class ReadDataRealData(SimulationRealData):
    def __init__(self, list_lines):
        """overwrite SimulationRealData __init__function"""
        self.list_lines_fit = list_lines


if __name__ == "__main__":
    if len(sys.argv) < 2:
        print("Please provide the name of the YAML file as an argument.")
        sys.exit(1)

    path_yaml_file = sys.argv[1]

    path_data_cloud, yaml_file = os.path.split(path_yaml_file)

    with open(os.path.abspath(f"{path_data_cloud}/{yaml_file}")) as f:
        params = yaml.safe_load(f)

    # for orionbar, the additive noise variance changes from line to line
    (
        _,  # df_int_fit
        _,  # y_fit
        sigma_a,  # _fit
        _,  # omega_fit
        _,  # y_valid
        _,  # sigma_a_valid
        _,  # omega_valid
    ) = ReadDataRealData(params["list_lines"]).setup_observation(
        data_int_path=f"{path_data_cloud}/{params['filename_int']}",
        data_err_path=f"{path_data_cloud}/{params['filename_err']}",
        save_obs=False,
    )

    approx_optim = ApproxParamsOptimNNBO(
        list_lines=params["list_lines"],
        sigma_a=sigma_a,
        sigma_m=np.log(params["sigma_m_float_linscale"]),
        **params["simu_init"],
    )
    approx_optim.main(**params["main_params"])



from typing import Optional

import numpy as np

from beetroots.sampler.saver.abstract_saver import Saver


class MySaver(Saver):
    def initialize_memory(
        self,
        T_MC: int,
        t: int,
        Theta: np.ndarray,
        forward_map_evals: dict = dict(),
        nll_utils: dict = dict(),
        dict_objective: dict = dict(),
        additional_sampling_log: dict = dict(),
    ) -> None:
        r"""initializes the memory with the correct shapes

        Parameters
        ----------
        T_MC : int
            size of the Markov chain to be sampled / optimization procedure
        t : int
            current iteration index
        Theta : np.ndarray
            current iterate in the Markov chain / optimization run
        forward_map_evals : dict, optional
            evaluations of the forward map and potentially derivatives, by default dict()
        nll_utils : dict, optional
            evaluation of utilitary values of the Likelihood class, by default dict()
        dict_objective : dict, optional
            contains the negative log posterior value and detailed components, by default dict()
        additional_sampling_log : dict, optional
            additional data on the sampling / optimization run, by default dict()
        """
        if self.batch_size is None:
            self.batch_size = T_MC

        self.t_last_init = t * 1
        self.next_batch_size = min(self.batch_size, (T_MC - t + 1) // self.freq_save)
        # print(t, self.next_batch_size)
        self.final_next_batch_size = self.next_batch_size

        self.memory["list_Theta"] = np.zeros(
            (self.final_next_batch_size, self.N, self.D_sampling)
        )

        if self.save_forward_map_evals:
            for k, v in forward_map_evals.items():
                if np.all(["grad" not in k, "hess_diag" not in k]):
                    self.memory[f"list_{k}"] = np.zeros(
                        (self.final_next_batch_size,) + v.shape
                    )

        for k, v in nll_utils.items():
            if np.all(["nll_" not in k, "grad" not in k, "hess_diag" not in k]):
                self.memory[f"list_{k}"] = np.zeros(
                    (self.final_next_batch_size,) + v.shape
                )

        for k, v in dict_objective.items():
            self.memory[f"list_{k}"] = np.zeros((self.final_next_batch_size,) + v.shape)

        for k, v in additional_sampling_log.items():
            if isinstance(v, np.ndarray):
                self.memory[f"list_{k}"] = np.zeros(
                    (self.final_next_batch_size,) + v.shape
                )
            else:
                self.memory[f"list_{k}"] = np.zeros(
                    (self.final_next_batch_size,),
                )

        self.memory["list_rng_state"] = np.zeros(
            (self.final_next_batch_size, 32),
            dtype=np.uint8,
        )
        self.memory["list_rng_inc"] = np.zeros(
            (self.final_next_batch_size, 32),
            dtype=np.uint8,
        )

    def update_memory(
        self,
        t: int,
        Theta: np.ndarray,
        forward_map_evals: dict = dict(),
        nll_utils: dict = dict(),
        dict_objective: dict = dict(),
        additional_sampling_log: dict = dict(),
        rng_state_array: Optional[np.ndarray] = None,
        rng_inc_array: Optional[np.ndarray] = None,
    ) -> None:
        r"""updates the memory with new information. All of the potential entries are optional except for the current iterate.

        Parameters
        ----------
        t : int
            current iteration index
        Theta : np.ndarray
            current iterate in the Markov chain / optimization run
        forward_map_evals : dict, optional
            evaluations of the forward map and potentially derivatives, by default dict()
        nll_utils : dict, optional
            evaluation of utilitary values of the Likelihood class, by default dict()
        dict_objective : dict, optional
            contains the negative log posterior value and detailed components, by default dict()
        additional_sampling_log : dict, optional
            additional data on the sampling / optimization run, by default dict()
        rng_state_array : Optional[np.ndarray], optional
            current state of the random generator (saved for sampling reproducibility), by default None
        rng_inc_array : Optional[np.ndarray], optional
            current inc of the random generator (saved for sampling reproducibility), by default None
        """
        t_save = (t - self.t_last_init) // self.freq_save

        Theta_full = np.zeros((Theta.shape[0], self.D))
        for i, idx in enumerate(self.list_idx_sampling):
            Theta_full[:, idx] = Theta[:, i]
        lin_Theta_full = self.scaler.from_scaled_to_lin(Theta_full)
        lin_Theta_full = lin_Theta_full[:, self.list_idx_sampling]
        self.memory["list_Theta"][t_save, :, :] = lin_Theta_full

        if self.save_forward_map_evals:
            for k, v in forward_map_evals.items():
                if np.all(["grad" not in k, "hess_diag" not in k]):
                    self.memory[f"list_{k}"][t_save] = v

        for k, v in nll_utils.items():
            if np.all(["nll_" not in k, "grad" not in k, "hess_diag" not in k]):
                self.memory[f"list_{k}"][t_save] = v

        for k, v in dict_objective.items():
            if k not in ["m_a", "s_a", "m_m", "s_m"]:
                self.memory[f"list_{k}"][t_save] = v

        for k, v in additional_sampling_log.items():
            self.memory[f"list_{k}"][t_save] = v

        if (rng_state_array is not None) and (rng_inc_array is not None):
            self.memory["list_rng_state"][t_save] = rng_state_array
            self.memory["list_rng_inc"][t_save] = rng_inc_array



import abc
from sys import byteorder
from typing import Optional

import numpy as np

from beetroots.modelling.posterior import Posterior
from beetroots.sampler.saver.abstract_saver import Saver
from beetroots.sampler.utils.mml import EBayesMMLE


class Sampler(abc.ABC):
    r"""abstract class designed to sample from a distribution defined with the ``Posterior`` class."""

    ESS_OPTIM = 1_000
    r"""number of random reproduced observations to draw to evaluate the model checking p-value for optimization procedures"""

    def __init__(
        self,
        D: int,
        L: int,
        N: int,
        rng: np.random.Generator = np.random.default_rng(42),
    ):
        r"""

        Parameters
        ----------
        D : int
            total number of physical parameters to reconstruct
        L : int
            number of observables per component :math:`n`
        N : int
            total number of pixels to reconstruct
        rng : np.random.Generator
            numpy random generator (for sampling reproducibility), by default np.random.default_rng(42)
        """
        self.D = D
        r"""int: total number of physical parameters to reconstruct"""
        self.L = L
        r"""int: number of observables per component :math:`n`"""
        self.N = N
        r"""int: total number of pixels to reconstruct"""

        self.rng = rng
        r"""np.random.Generator: numpy random generator (for sampling reproducibility)"""

    def get_rng_state(self):
        r"""extracts the current state and inc of the random generator

        Returns
        -------
        rng_state_array : np.array of ints
            current state of the random generator
        rng_inc_array : np.array of ints
            current inc of the random generator

        Note
        ----
        1. state and inc are very large integers, and thus need to be converted to hex format (later to an array of ints) to be saved in a .h5 file https://docs.python.org/3/library/stdtypes.html#int.to_bytes
        2. need 32 bytes in length: otherwise, the inverse operation int.from_bytes(state_array,byteorder) does not coincide with the original int value
        """
        rng_state_array = np.array(
            bytearray(
                self.rng.__getstate__()["state"]["state"].to_bytes(32, byteorder),
            )
        )
        rng_inc_array = np.array(
            bytearray(
                self.rng.__getstate__()["state"]["inc"].to_bytes(32, byteorder),
            )
        )
        return rng_state_array, rng_inc_array

    def set_rng_state(self, state_bytes, inc_bytes):
        r"""sets the state and inc of the random generator to a specific value. Enables to re-launch a sampling when the rng state is regularly saved (see ``MySaver`` class).

        Parameters
        ----------
        state_bytes : bytes
            state to be set
        inc_bytes : bytes
            inc to be set
        """
        # step 1 : convert bytes to int
        loaded_state = int.from_bytes(state_bytes, byteorder)
        loaded_inc = int.from_bytes(inc_bytes, byteorder)

        # step 2 : set rng state
        state = self.rng.__getstate__()
        state["state"]["state"] = loaded_state
        state["state"]["inc"] = loaded_inc
        self.rng.__setstate__(state)

    def generate_random_start_Theta(self, posterior: Posterior):
        r"""generates a random element of the hypercube defined by the lower and upper bounds with uniform distribution

        Parameters
        ----------
        posterior : Posterior
            contains the lower and upper bounds of the hypercube

        Returns
        -------
        Theta : np.array of shape (N, D)
            random element of the hypercube defined by the lower and upper bounds with uniform distribution
        """
        if posterior.prior_indicator is not None:
            Theta = (
                self.rng.uniform(0, 1, size=(posterior.N, posterior.D))
                * (
                    posterior.prior_indicator.upper_bounds[None, :]
                    - posterior.prior_indicator.lower_bounds[None, :]
                )
                + posterior.prior_indicator.lower_bounds[None, :]
            )

        else:
            Theta = self.rng.standard_normal(size=(posterior.N, posterior.D))

        return Theta

    @abc.abstractmethod
    def generate_random_start_Theta_1pix(
        self, Theta: np.ndarray, posterior: Posterior, idx_pix: np.ndarray
    ) -> np.ndarray:
        r"""draws a random vectors for components :math:`n` (e.g., a pixel :math:`\theta_n`).

        Parameters
        ----------
        Theta : np.ndarray
            current iterate
        posterior : Posterior
            contains the lower and upper bounds of the hypercube
        idx_pix : np.ndarray
            indices of the pixels

        Returns
        -------
        np.array of shape (n_pix, self.k_mtm, D)
            random element of the hypercube defined by the lower and upper bounds with uniform distribution
        """
        pass

    @abc.abstractmethod
    def sample(
        self,
        posterior: Posterior,
        saver: Saver,
        max_iter: int,
        Theta_0: Optional[np.ndarray],
        T_BI: Optional[int],
        **kwargs,
    ) -> None:
        """main method of the class, runs the sampler on the input posterior distribution

        Parameters
        ----------
        posterior : Posterior
            probability distribution to be sampled
        saver : Saver
            object responsible for progressively saving the Markov chain data during the run
        max_iter : int
            total duration of a Markov chain
        Theta_0 : Optional[np.ndarray], optional
            starting point
        T_BI : int, optional
            duration of the `Burn-in` phase
        """
        pass

    def sample_regu_hyperparams(
        self,
        posterior: Posterior,
        regu_weights_optimizer: EBayesMMLE,
        t: int,
        current_Theta: np.ndarray,
    ) -> np.ndarray:
        r"""updates the spatial regularization weight vector (denoted :math:`\tau` in :cite:t:`paludEfficientSamplingNon2023`) using the marginal likelihood optimizer defined in :cite:t:`vidalMaximumLikelihoodEstimation2020`

        Parameters
        ----------
        posterior : Posterior
            posterior distribution
        regu_weights_optimizer : EBayesMMLE
            marginal likelihood optimizer
        t : int
            time step
        current_Theta : np.ndarray
            current iterate

        Returns
        -------
        np.ndarray
            updated spatial regularization weight vector
        """
        assert self.N > 1
        assert posterior.prior_spatial is not None

        tau_tm1 = posterior.prior_spatial.weights * 1

        neglog_prior = posterior.prior_spatial.neglog_pdf(
            current_Theta,
            with_weights=False,
        )  # (D,)

        new_tau = regu_weights_optimizer.update(tau_tm1, t, neglog_prior)
        # new_tau[3] = 20  # fix tau of A_v
        return new_tau



import abc
import datetime
import os
import sys
import warnings
from typing import Dict, List, Optional

import matplotlib.pyplot as plt
import yaml
from cerberus import Validator

from beetroots.inversion.reporting.report_astro import ReportAstro

warnings.filterwarnings("ignore")


class Simulation(abc.ABC):
    r"""abstract class for the main class of the inversion. Its instanciations set up, launch and analyze the results of inversions."""

    def __init__(
        self,
        simu_name: str,
        cloud_name: str,
        params_names: Dict[str, str],
        list_lines_fit: List[str],
        params: dict,
        max_workers: int = 10,
        small_size: int = 16,
        medium_size: int = 20,
        bigger_size: int = 24,
        folder_path: Optional[str] = None,
    ):
        r"""

        Parameters
        ----------
        simu_name : str
            name of the full inversion procedure, used to name the outputs folder
        cloud_name : str
            name of the observed cloud
        params_names : Dict[str, str]
            pairs of names for each parameter, with first the standard name (to be found as title of column in DataFrames) and second a latex name (to be displayed in figures). For instance, for the thermal pressure "P": r"$P_{th}$)"
        list_lines_fit : List[str]
            names of the observables used for the inversion
        params : dict
            content of the ``.yaml`` input file
        max_workers : int, optional
            maximum number of workers that can be used for inversion or results extraction, by default 10
        small_size : int, optional
            size for basic text, axes titles, xticks and yticks, by default 16
        medium_size : int, optional
            size of the axis labels, by default 20
        bigger_size : int, optional
            size of the figure title, by default 24
        folder_path : Optional[str], optional
            path to the output folder, by default None
        """
        self.cloud_name = cloud_name
        r"""str: name of the observed cloud"""
        self.list_lines_fit = list_lines_fit
        r"""List[str]: names of the observables used for inversion"""
        self.L = len(list_lines_fit)
        r"""int: number of observables per component :math:`n` used for inversion, e.g., per pixel"""
        self.simu_name = simu_name
        r"""str: name of the full inversion procedure, used to name the outputs folder"""

        self.create_empty_output_folders(simu_name, params, folder_path)
        self.setup_plot_text_sizes(small_size, medium_size, bigger_size)

        self.max_workers = max_workers
        r"""int: maximum number of workers that can be used for inversion or results extraction"""

        self.list_names = list(params_names.keys())
        r"""List[str]: names of the physical parameters in files, e.g., as titles of a DataFrame column, e.g., P_th or P for the thermal pressure"""
        self.list_names_plots = list(params_names.values())
        r"""List[str]: names of the physical parameters in figures, e.g., $P_{th}$ for the thermal pressure"""

        self.D = len(self.list_names)  # Number of physical parameters
        r"""int: total number of physical parameters"""

        self.list_fixed_values = list(params["forward_model"]["fixed_params"].values())
        r"""List[float | None] list of values for the physical parameters that are not to be samples"""

        try:
            pixels_of_interest = params["pixels_of_interest"]
        except:
            print("no pixels of interest")
            pixels_of_interest = {}
        self.pixels_of_interest = pixels_of_interest
        r"""Dict[int, str]: pairs of (index, name) for pixels of interest that are to be highlighted in maps"""

        self.list_idx_sampling = [
            i
            for i, v in enumerate(params["forward_model"]["fixed_params"].values())
            if v is None
        ]
        r"""List[int]: indices of the physical parameters that are to be sampled (the other ones are fixed)"""

        self.D_sampling = len(self.list_idx_sampling)
        r"""int: number of physical parameters to be sampled"""

        # number of params used in forward model
        if "kappa" in self.list_names:
            D_no_kappa = self.D - 1
        else:
            D_no_kappa = self.D * 1
        self.D_no_kappa = D_no_kappa * 1
        r"""number of physical parameters excluding the scaling parameter :math:`\kappa` (used in astrophysical applications). When "kappa" is not in ``list_names``, then ``D_no_kappa = D``"""

    def setup_plot_text_sizes(
        self,
        small_size: int = 16,
        medium_size: int = 20,
        bigger_size: int = 24,
    ) -> None:
        r"""defines text sizes on matplotlib plots

        Parameters
        ----------
        small_size : int, optional
            size for basic text, axes titles, xticks and yticks, by default 16
        medium_size : int, optional
            size of the axis labels, by default 20
        bigger_size : int, optional
            size of the figure title, by default 24
        """
        plt.rc("font", size=small_size)  # controls default text sizes
        plt.rc("axes", titlesize=small_size)  # fontsize of the axes title
        plt.rc("axes", labelsize=medium_size)  # fontsize of the x and y labels
        plt.rc("xtick", labelsize=small_size)  # fontsize of the tick labels
        plt.rc("ytick", labelsize=small_size)  # fontsize of the tick labels
        plt.rc("legend", fontsize=small_size)  # legend fontsize
        plt.rc("figure", titlesize=bigger_size)  # fontsize of the figure title
        return

    def create_empty_output_folders(
        self, name: str, params: dict, folder_path: Optional[str] = None
    ) -> None:
        r"""creates the directories that receive the results of the sampling, and saves the ``.yaml`` input file there for reproducibility

        Parameters
        ----------
        name : str
            name of the simulation to be run
        params : dict
            set of params used for the simulation
        folder_path: str, optional
            folder where to write outputs, by default None

        Returns
        -------
        tuple of str
            paths of the created directories
        """
        now = datetime.datetime.now()
        dt_string = now.strftime("%Y-%m-%d_%H")

        # path to the outputs dir
        if folder_path is None:
            path_ouput_general = f"{os.path.abspath(__file__)}/../../../outputs"
        else:
            path_ouput_general = f"{folder_path}/outputs"

        path_ouput_general = os.path.abspath(path_ouput_general)

        path_output_sim = f"{path_ouput_general}/{name}_{dt_string}"
        self.path_output_sim = os.path.abspath(path_output_sim)
        r"""str: path to the root folder of the inversion results, e.g., ``./outputs/simu1``"""

        self.path_img = path_output_sim + "/img"
        r"""str: path to the image folder in the inversion results, e.g., ``./outputs/simu1/img``"""
        self.path_raw = path_output_sim + "/raw"
        r"""str: path to the folder containing the raw inversion results, i.e., the ``.hdf5`` files, e.g., ``./outputs/simu1/raw``"""
        self.path_data_csv = path_output_sim + "/data"
        r"""str: path to the data folder in the inversion results, e.g., ``./outputs/simu1/data``"""
        self.path_data_csv_in = self.path_data_csv + "/input"
        r"""str: path to the input data folder in the inversion results, e.g., ``./outputs/simu1/data/input``"""
        self.path_data_csv_out = self.path_data_csv + "/output"
        r"""str: path to the output data folder in the inversion results, e.g., ``./outputs/simu1/data/output``"""
        self.path_data_csv_out_mcmc = self.path_data_csv_out + "/mcmc"
        r"""str: path to the mcmc output data folder in the inversion results, e.g., ``./outputs/simu1/data/output/mcmc``"""
        self.path_data_csv_out_optim_map = self.path_data_csv_out + "/optim_map"
        r"""str: path to the MAP estimation (with optimization approach) output data folder in the inversion results, e.g., ``./outputs/simu1/data/output/optim_map``"""
        self.path_data_csv_out_optim_mle = self.path_data_csv_out + "/optim_mle"
        r"""str: path to the MLE estimation (with optimization approach) output data folder in the inversion results, e.g., ``./outputs/simu1/data/output/optim_mle``"""

        # create the folders if necessary
        for folder_path in [
            path_ouput_general,
            self.path_output_sim,
            self.path_img,
            self.path_raw,
            self.path_data_csv,
            self.path_data_csv_in,
            self.path_data_csv_out,
            self.path_data_csv_out_mcmc,
            self.path_data_csv_out_optim_map,
            self.path_data_csv_out_optim_mle,
        ]:
            if not os.path.isdir(folder_path):
                os.mkdir(folder_path)

        # empty potentially existing results
        for folder_path in [
            self.path_data_csv_out_mcmc,
            self.path_data_csv_out_optim_map,
            self.path_data_csv_out_optim_mle,
        ]:
            for filename in os.listdir(folder_path):
                file_path = os.path.join(folder_path, filename)
                if os.path.isfile(file_path) or os.path.islink(file_path):
                    os.unlink(file_path)

        # equivalent of copy paste of the config file to output folder
        with open(f"{self.path_data_csv_in}/config_{name}.yaml", "w") as file:
            yaml.dump(params, file)

        return

    @classmethod
    def load_params(cls, path_data_cloud: str) -> dict:
        r"""

        Parameters
        ----------
        path_data_cloud : str
            path to the folder containing the ``.yaml`` input file. This folder should also contain the data necessary to set up the inversion.

        Returns
        -------
        dict
            content of the ``.yaml`` input file
        """
        # print("don't forget to put back sys.argv in load_params")
        if len(sys.argv) < 2:
            print("Please provide the name of the YAML file as an argument.")
            sys.exit(1)

        yaml_file = sys.argv[1]

        # ? Set a filename here for debugging.
        # yaml_file = "input_params_nn_direct_N10_fixed_angle_nopmala.yaml"

        with open(os.path.abspath(f"{path_data_cloud}/{yaml_file}")) as f:
            params = yaml.safe_load(f)

        return params

    @classmethod
    def check_input_params_file(cls, params: dict, schema: dict) -> None:
        # inputs validation
        v = Validator(schema, allow_unknown=True)
        is_input_correct: bool = v.validate(params)

        print(f"is input file correct: {is_input_correct}")

        if not is_input_correct:
            print(f"identified errors: {v.errors}")

        return None

    def generate_report(
        self,
        list_model_names: List[str],
        to_run_optim_map: bool,
        to_run_mcmc: bool,
    ):
        r"""gathers the results of the inversion, such as images, in a single ``.pdf`` file to simplify the result understanding

        .. warning::

            Unfinished method.
            Should only be used for astrophysical applications.

        Parameters
        ----------
        list_model_names : List[str]
            names of the models to extract. An inversion relies on one model only.
        to_run_optim_map : bool
            wether the MAP estimation with an optimization procedure was run, and thus wether the report should display the associated results
        to_run_mcmc : bool
            wether an MCMC-based inference was run, and thus wether the report should display the associated results
        """
        # Create a new report object
        my_report = ReportAstro(
            self.D,
            self.D_sampling,
            self.path_img,
            list_model_names,
            self.pixels_of_interest,
            to_run_optim_map,
            to_run_mcmc,
        )
        my_report.main(
            self.cloud_name,
            self.simu_name,
            self.list_lines_fit,
            self.list_lines_valid,
        )



from typing import Dict, List, Tuple

import numpy as np
import pandas as pd

from beetroots.inversion.plots.map_shaper import MapShaper
from beetroots.simulations.astro.observation.abstract_observation import (
    SimulationObservation,
)


class SimulationRealData(SimulationObservation):
    def setup_observation(
        self,
        data_int_path: str,
        data_err_path: str,
        save_obs: bool = True,
    ) -> Tuple[pd.DataFrame, np.ndarray, np.ndarray, np.ndarray]:
        # * read observations
        df_int = pd.read_pickle(data_int_path)
        df_err = pd.read_pickle(data_err_path)
        assert list(df_int.index.names) == ["X", "Y"]
        assert list(df_err.index.names) == ["X", "Y"]
        assert len(df_int) == len(df_err)

        df_int = df_int.sort_index()
        df_err = df_err.sort_index()

        if "idx" not in list(df_err.columns):
            df_int["idx"] = np.arange(len(df_int))
            df_err["idx"] = np.arange(len(df_err))

        assert df_int.shape == df_err.shape
        assert list(df_int.columns) == list(df_err.columns)

        for line in self.list_lines_fit:
            assert line in list(df_int.columns)

        self.list_lines_valid = list(
            set(list(df_int.columns)) - set(self.list_lines_fit)
        )
        self.list_lines_valid.remove("idx")
        self.list_lines_valid.sort()

        df_int_fit = df_int.loc[:, self.list_lines_fit + ["idx"]]
        df_err_fit = df_err.loc[:, self.list_lines_fit + ["idx"]]

        df_int_valid = df_int.loc[:, self.list_lines_valid + ["idx"]]
        df_err_valid = df_err.loc[:, self.list_lines_valid + ["idx"]]

        # eliminate absurd values
        for col in self.list_lines_fit:
            df_int_fit[col] = df_int_fit[col].apply(
                lambda x: x if x <= 1e3 else np.inf,
            )
            df_err_fit[col] = df_err_fit[col].apply(
                lambda x: x if x <= 1e3 else np.inf,
            )
        for col in self.list_lines_valid:
            df_int_valid[col] = df_int_valid[col].apply(
                lambda x: x if x <= 1e3 else np.inf,
            )
            df_err_valid[col] = df_err_valid[col].apply(
                lambda x: x if x <= 1e3 else np.inf,
            )

        self.N = len(df_int_fit)

        # * correct values (censoring / errors / etc.)
        df_censor_fit = df_err_fit.copy()
        # df_censor.iloc[:, :-1] *= 3 # with potential censorship
        df_censor_fit.iloc[:, :-1] = 1e-60  # no censorship

        df_censor_valid = df_err_valid.copy()
        # df_censor.iloc[:, :-1] *= 3 # with potential censorship
        df_censor_valid.iloc[:, :-1] = 1e-60  # no censorship

        y_fit = np.nan_to_num(df_int_fit.drop("idx", 1).values, nan=1e-15)
        sigma_a_fit = np.nan_to_num(df_err_fit.drop("idx", 1).values, nan=1)
        omega_fit = df_censor_fit.drop("idx", 1).values

        y_valid = np.nan_to_num(df_int_valid.drop("idx", 1).values, nan=1e-15)
        sigma_a_valid = np.nan_to_num(
            df_err_valid.drop("idx", 1).values,
            nan=1,
        )
        omega_valid = df_censor_valid.drop("idx", 1).values

        self.Theta_true_scaled = None
        self.map_shaper = MapShaper(df_int)

        # * save observation
        if save_obs:
            df_int_fit.to_pickle(
                f"{self.path_data_csv_in}/observation_maps.pkl",
            )
            df_err_fit.to_pickle(f"{self.path_data_csv_in}/additive_std.pkl")
            df_censor_fit.to_pickle(
                f"{self.path_data_csv_in}/censor_threshold.pkl",
            )

        return (
            df_int_fit,
            y_fit,
            sigma_a_fit,
            omega_fit,
            y_valid,
            sigma_a_valid,
            omega_valid,
        )



import abc

import matplotlib.pyplot as plt
from matplotlib import patches


class AbstractPlots2D(abc.ABC):
    """abstract class for drawing figures of maps of observations, noise standard deviations, estimated physical parameters, estimated credibility interval sizes, etc."""

    def _draw_rect_on_pixels_of_interest(self):
        """Draws a rectangle around the pixels of interest. Requires the class to have a `pixels_of_interest_coords` attribute, as a Dict with the vector index as keys and the corresponding map coordinates as values."""
        ax = plt.gca()

        for _, coords in self.pixels_of_interest_coords.items():
            rect = patches.Rectangle(
                (coords[0] - 0.5, coords[1] - 0.5),
                1,
                1,
                linewidth=1,
                edgecolor="k",
                facecolor="none",
            )
            ax.add_patch(rect)

        return



"""Essentially contains the definition of

1. The forward model (an polynomial regression of a grid of simulations of the Meudon PDR code)
2. the negative logpdf of the likelihood, prior and posterior, their gradient and the diagonal of their hessian matrix
"""



# import json
import os
import pickle
import warnings
from typing import Dict, Optional, Tuple

from beetroots.modelling.forward_maps.regression_poly import PolynomialApprox
from beetroots.simulations.astro.forward_map.abstract_forward_map import (
    SimulationForwardMap,
)
from beetroots.space_transform.transform import MyScaler

warnings.filterwarnings("ignore")


class SimulationPolynomialReg(SimulationForwardMap):
    r"""abstract class for to set up the forward map to an already defined polynomial for an inversion of astrophysical data"""

    def setup_forward_map(
        self,
        forward_model_name: str,
        angle: float,
        dict_fixed_params: Dict[str, Optional[float]],
        dict_is_log_scale_params: Dict[str, bool],
    ) -> Tuple[MyScaler, PolynomialApprox]:

        with open(
            f"{self.MODELS_PATH}/{forward_model_name}/scaler.pickle",
            "rb",
        ) as file_:
            scaler_sklearn = pickle.load(file_)

        scaler = MyScaler(
            mean_=scaler_sklearn.mean_[:-1].flatten(),
            std_=scaler_sklearn.scale_[:-1].flatten(),
            list_is_log=list(dict_is_log_scale_params.values()),
        )

        # transformation from linear scale (in degrees) to scaled
        angle_scaled = (angle - 30.0) / 20.0

        dict_fixed_params_scaled = self.scale_dict_fixed_params(
            scaler, dict_fixed_params
        )

        # load forward model
        forward_map = PolynomialApprox(
            forward_model_name,
            dict_fixed_params_scaled,
            angle_scaled,
        )
        forward_map.restrict_to_output_subset(self.list_lines_fit)
        return scaler, forward_map



import os
from typing import Dict, List

from pylatex import (
    Command,
    Document,
    Figure,
    Itemize,
    Label,
    NewPage,
    NoEscape,
    Package,
    Section,
    Subsection,
)

from beetroots.inversion.plots.readable_line_names import lines_to_latex


class ReportAstro:
    author = "Pierre Palud"

    def __init__(
        self,
        D: int,
        D_sampling: int,
        path_folder_img: str,
        list_model_names: List[str],
        pixels_of_interest: Dict[int, str],
        to_run_optim_map: bool,
        to_run_mcmc: bool,
    ):
        self.doc = Document()

        self.doc.packages.append(Package("float"))
        self.doc.packages.append(Package("hyperref"))

        self.D = D
        self.D_sampling = D_sampling

        self.path_folder_img = path_folder_img
        self.list_model_names = list_model_names
        self.pixels_of_interest = pixels_of_interest

        self.to_run_optim_map = to_run_optim_map
        self.to_run_mcmc = to_run_mcmc

    def add_title(
        self,
        cloud_name: str,
        simu_name: str,
    ) -> None:
        self.doc.preamble.append(
            Command("title", NoEscape(simu_name.replace("_", " ")))
        )
        # self.doc.preamble.append(Command("subtitle", simu_name))
        self.doc.preamble.append(Command("author", self.author))
        # self.doc.preamble.append(Command("date", dt_string))
        self.doc.append(Command("maketitle"))

        self.doc.append(NoEscape("\\tableofcontents"))
        self.doc.append(NewPage())
        return

    def add_section_observations(
        self,
        list_lines: List[str],
        list_lines_valid: List[str],
    ) -> None:
        path_folder_obs = f"{self.path_folder_img}/observations"

        L = len(list_lines)
        ncols = 5 if L > 5 else L
        nrows = L // ncols + (L % ncols > 0)
        width = 0.98 * (1 / ncols)

        with self.doc.create(
            Section(
                "Observations",
                label=False,
            )
        ):

            with self.doc.create(
                Subsection(
                    "Observation maps",
                    label=False,
                )
            ):
                # label_figure_log = r"fig:obs_log"
                # self.doc.append(
                #     r"Fig.~\ref{"
                #     + label_figure_log
                #     + r"} shows the observation maps used for inversion in log scale."
                # )
                # label_figure_lin = r"fig:obs_lin"
                # self.doc.append(
                #     r"Fig.~\ref{"
                #     + label_figure_lin
                #     + r"} shows the same observation maps in linear scale."
                # )

                with self.doc.create(Figure(position="H")) as image:
                    ell = 0
                    for i in range(nrows):
                        for j in range(ncols):
                            if ell < L:
                                if ell % ncols == 0 and ell > 0:
                                    image.append(NoEscape(r"\\"))

                                image.add_image(
                                    f"{path_folder_obs}/observation_line_{ell}_{list_lines[ell]}.PNG",
                                    width=NoEscape(f"{width}\\textwidth"),
                                    placement=None,
                                )
                                ell += 1

                    image.add_caption("Observation maps (log scale).")
                    # image.append(Label(label_figure_log))

                with self.doc.create(Figure(position="H")) as image:
                    ell = 0
                    for i in range(nrows):
                        for j in range(ncols):
                            if ell < L:
                                if ell % ncols == 0 and ell > 0:
                                    image.append(NoEscape(r"\\"))

                                image.add_image(
                                    f"{path_folder_obs}/observation_line_{ell}_{list_lines[ell]}_linscale.PNG",
                                    width=NoEscape(f"{width}\\textwidth"),
                                    placement=None,
                                )
                                ell += 1
                    image.add_caption("Observation maps (linear scale).")
                    # image.append(Label(label_figure_lin))

            with self.doc.create(
                Subsection(
                    "Standard deviation maps",
                    label=False,
                )
            ):

                with self.doc.create(Figure(position="H")) as image:
                    ell = 0
                    for i in range(nrows):
                        for j in range(ncols):
                            if ell < L:
                                if ell % ncols == 0 and ell > 0:
                                    image.append(NoEscape(r"\\"))

                                image.add_image(
                                    f"{path_folder_obs}/add_err_std_line_{ell}_{list_lines[ell]}.PNG",
                                    width=NoEscape(f"{width}\\textwidth"),
                                    placement=None,
                                )
                                ell += 1

                    image.add_caption("Maps of additive noise standard deviation.")

            with self.doc.create(
                Subsection(
                    "Censorship maps",
                    label=False,
                )
            ):

                with self.doc.create(Figure(position="H")) as image:
                    ell = 0
                    for i in range(nrows):
                        for j in range(ncols):
                            if ell < L:
                                if ell % ncols == 0 and ell > 0:
                                    image.append(NoEscape(r"\\"))

                                image.add_image(
                                    f"{path_folder_obs}/censored_map_line_{ell}_{list_lines[ell]}.PNG",
                                    width=NoEscape(f"{width}\\textwidth"),
                                    placement=None,
                                )
                                ell += 1

                    image.add_image(
                        f"{path_folder_obs}/proportion_censored_lines.PNG",
                        width=NoEscape(f"{width}\\textwidth"),
                        placement=None,
                    )

                    image.add_caption("Maps of censorship.")

            with self.doc.create(
                Subsection(
                    "Signal to Noise Ratio maps",
                    label=False,
                )
            ):

                with self.doc.create(Figure(position="H")) as image:
                    ell = 0
                    for i in range(nrows):
                        for j in range(ncols):
                            if ell < L:
                                if ell % ncols == 0 and ell > 0:
                                    image.append(NoEscape(r"\\"))

                                image.add_image(
                                    f"{path_folder_obs}/snr_line_{ell}_{list_lines[ell]}.PNG",
                                    width=NoEscape(f"{width}\\textwidth"),
                                    placement=None,
                                )
                                ell += 1

                    image.add_caption("Maps of SNR.")

        self.doc.append(NewPage())
        return

    def add_section_model(self, model_name: str) -> None:
        with self.doc.create(
            Section(
                NoEscape(model_name.replace("_", " ")),
                label=False,
            )
        ):
            self.add_subsection_objective(model_name)
            self.add_subsection_accepted_freq(model_name)

            self.add_subsection_pvalues(model_name)
            self.add_subsection_estimators(model_name)
            self.add_subsection_pixels_of_interest(model_name)

        # self.doc.append(NewPage())
        return

    def add_subsection_pvalues(
        self,
        model_name: str,
    ):
        path_folder_p = f"{self.path_folder_img}/p-values/{model_name}"

        with self.doc.create(
            Subsection(
                NoEscape("p-values maps"),
                label=False,
            )
        ):

            with self.doc.create(Figure(position="H")) as image:
                image.add_image(
                    f"{path_folder_p}/pval_estim_mcmc_seed0.PNG",
                    width=NoEscape("0.45\\textwidth"),
                )
                image.add_image(
                    f"{path_folder_p}/decision_pval_estim_mcmc_seed0.PNG",
                    width=NoEscape("0.45\\textwidth"),
                )
                if os.path.isfile(f"{path_folder_p}/proba_reject_mcmc_seed0.PNG"):
                    image.append(NoEscape(r"\\"))

                    image.add_image(
                        f"{path_folder_p}/proba_reject_mcmc_seed0.PNG",
                        width=NoEscape("0.45\\textwidth"),
                    )
                    image.add_image(
                        f"{path_folder_p}/decision_from_bayes_pval_mcmc_seed0.PNG",
                        width=NoEscape("0.45\\textwidth"),
                    )

                image.add_caption("Results of the p-values analysis.")
        return

    def add_subsection_estimators(
        self,
        model_name: str,
    ):
        path_folder_true = f"{self.path_folder_img}/true"
        path_folder_est = f"{self.path_folder_img}/estimators/{model_name}"

        ncols = 4 if os.path.isdir(path_folder_true) else 3
        width = 0.98 * (1 / ncols)

        with self.doc.create(
            Subsection(
                "Reconstructed physical parameters maps",
                label=False,
            )
        ):

            with self.doc.create(Figure(position="H")) as image:
                for d in range(self.D):
                    if os.path.isfile(f"{path_folder_est}/MAP_mcmc/MAP_mcmc_{d}.PNG"):

                        if os.path.isdir(path_folder_true):
                            image.add_image(
                                f"{path_folder_true}/true_{d}.PNG",
                                width=NoEscape(f"{width}\\textwidth"),
                                placement=None,
                            )
                        image.add_image(
                            f"{path_folder_est}/MAP_mcmc/MAP_mcmc_{d}.PNG",
                            width=NoEscape(f"{width}\\textwidth"),
                            placement=None,
                        )
                        image.add_image(
                            f"{path_folder_est}/MMSE/MMSE_{d}.PNG",
                            width=NoEscape(f"{width}\\textwidth"),
                            placement=None,
                        )

                        filename = (
                            f"{path_folder_est}/CI68/68_CI_uncertainty_factor_d{d}.PNG"
                        )
                        if os.path.isfile(filename):
                            image.add_image(
                                filename,
                                width=NoEscape(f"{width}\\textwidth"),
                                placement=None,
                            )
                        image.append(NoEscape(r"\\"))

                image.add_caption("Reconstructed physical parameters maps (log scale).")

            with self.doc.create(Figure(position="H")) as image:
                for d in range(self.D):
                    if os.path.isdir(path_folder_true):
                        image.add_image(
                            f"{path_folder_true}/true_linscale_{d}.PNG",
                            width=NoEscape(f"{width}\\textwidth"),
                            placement=None,
                        )
                    image.add_image(
                        f"{path_folder_est}/MAP_mcmc/MAP_mcmc_linscale_{d}.PNG",
                        width=NoEscape(f"{width}\\textwidth"),
                        placement=None,
                    )
                    image.add_image(
                        f"{path_folder_est}/MMSE/MMSE_linscale_{d}.PNG",
                        width=NoEscape(f"{width}\\textwidth"),
                        placement=None,
                    )
                    filename = f"{path_folder_est}/CI68/68_CI_uncertainty_factor_linscale_d{d}.PNG"
                    if os.path.isfile(filename):
                        image.add_image(
                            filename,
                            width=NoEscape(f"{width}\\textwidth"),
                            placement=None,
                        )
                    image.append(NoEscape(r"\\"))

                image.add_caption(
                    "Reconstructed physical parameters maps (linear scale)."
                )

        return

    def add_subsection_pixels_of_interest(
        self,
        model_name: str,
    ):
        path_folder_hist = f"{self.path_folder_img}/mc/{model_name}_2D/hist"
        path_folder_proba = f"{self.path_folder_img}/mc/{model_name}_2D/proba_contours"

        path_folder_yfx = f"{self.path_folder_img}/distri_comp_yspace/{model_name}"

        width = 0.95 / (self.D_sampling - 1)

        for pix_idx, name in self.pixels_of_interest.items():
            pix_text = f" for pixel {name} (idx={pix_idx})"

            with self.doc.create(
                Subsection(
                    f"Pixel {name} (idx={pix_idx})",
                    label=False,
                )
            ):

                with self.doc.create(Figure(position="H")) as image:
                    for d2 in range(self.D - 1, 0, -1):
                        tot = 0
                        for d1 in range(0, d2):
                            filename = f"{path_folder_hist}/hist2D_n{pix_idx}_d1{d1}_d2{d2}_overall_chain.PNG"
                            if os.path.isfile(filename):
                                image.add_image(
                                    filename,
                                    width=NoEscape(f"{width}\\textwidth"),
                                    placement=None,
                                )
                                tot += 1

                        if tot > 0:
                            image.append(NoEscape(r"\\"))

                    image.add_caption(
                        NoEscape(
                            r"2D Histograms of inferred physical parameters" + pix_text
                        ),
                    )

                with self.doc.create(Figure(position="H")) as image:
                    for d2 in range(self.D - 1, 0, -1):
                        tot = 0
                        for d1 in range(0, d2):
                            filename = f"{path_folder_proba}/HPR_2D_n{pix_idx}_d1{d1}_d2{d2}_overall_chain.PNG"
                            if os.path.isfile(filename):
                                image.add_image(
                                    filename,
                                    width=NoEscape(f"{width}\\textwidth"),
                                    placement=None,
                                )
                                tot += 1
                        if tot > 0:
                            image.append(NoEscape(r"\\"))

                    image.add_caption(
                        NoEscape(
                            r"2D contours of High Probability regions of inferred physical parameters"
                            + pix_text
                        ),
                    )

                with self.doc.create(Figure(position="H")) as image:
                    image.add_image(
                        f"{path_folder_yfx}/distribution_comparison_pix_{pix_idx}_fit.PNG",
                        width=NoEscape("0.45\\textwidth"),
                    )
                    image.add_image(
                        f"{path_folder_yfx}/distribution_comparison_pix_{pix_idx}_valid.PNG",
                        width=NoEscape("0.45\\textwidth"),
                    )
                    image.add_caption(
                        NoEscape(
                            r"Comparison of distributions on $y$ and $f(\theta)$"
                            + pix_text
                        ),
                    )

    def add_subsection_objective(
        self,
        model_name: str,
    ):
        path_folder_obj = f"{self.path_folder_img}/objective/{model_name}"
        with self.doc.create(
            Subsection(
                "Objective",
                label=False,
            )
        ):
            # label_figure = f"fig:objective_{model_name}"
            # self.doc.append(
            #     r"Fig.~\ref{"
            #     + label_figure
            #     + r"} shows the evolution of objective during sampling."
            # )

            with self.doc.create(Figure(position="H")) as image:
                image.add_image(
                    f"{path_folder_obj}/mcmc_objective_with_bi_and_true.PNG",
                    width=NoEscape("0.45\\textwidth"),
                )
                image.add_image(
                    f"{path_folder_obj}/mcmc_objective_no_bi_with_true.PNG",
                    width=NoEscape("0.45\\textwidth"),
                )
                image.add_caption("Evolution of objective during sampling.")
                # image.append(Label(label_figure))
        return

    def add_subsection_accepted_freq(
        self,
        model_name,
    ) -> None:
        path_folder_obj = f"{self.path_folder_img}/accepted_freq/mcmc/{model_name}"

        with self.doc.create(
            Subsection(
                "Accept frequency",
                label=False,
            )
        ):

            with self.doc.create(Figure(position="H")) as image:
                image.add_image(
                    f"{path_folder_obj}/freq_accept_seed0_MTM.PNG",
                    width=NoEscape("0.45\\textwidth"),
                )
                image.add_image(
                    f"{path_folder_obj}/freq_accept_seed0_PMALA.PNG",
                    width=NoEscape("0.45\\textwidth"),
                )
                image.add_caption("Evolution of accept frequency during sampling.")

        return

    def main(
        self,
        cloud_name: str,
        simu_name: str,
        list_lines: List[str],
        list_lines_valid: List[str],
    ) -> None:
        print("\nstarting report...")
        self.add_title(cloud_name, simu_name)

        self.add_section_observations(list_lines, list_lines_valid)

        for model_name in self.list_model_names:
            self.add_section_model(model_name)

        self.doc.generate_pdf(
            f"{self.path_folder_img}/../report",
            clean_tex=False,
        )
        print("report exported to pdf.")
        return



"""Defines a Dataclass that stores the parameters of the augmented PSGLD sampler
"""
from typing import List, Union

import numpy as np
import pandas as pd


class MySamplerParams(object):
    r"""Dataclass that stores the parameters of the sampler proposed in :cite:t:`paludEfficientSamplingNon2023`"""

    __slots__ = (
        "initial_step_size",
        "extreme_grad",
        "history_weight",
        "selection_probas",
        "k_mtm",
        "is_stochastic",
        "compute_correction_term",
    )

    def __init__(
        self,
        initial_step_size: float,
        extreme_grad: float,
        history_weight: float,
        selection_probas: Union[np.ndarray, List[float]],
        k_mtm: int,
        is_stochastic: bool = True,
        compute_correction_term: bool = True,
    ) -> None:
        r"""

        Parameters
        ----------
        initial_step_size : float
            step size used in the Position-dependent MALA transition kernel, denoted :math:`\epsilon` in the article
        extreme_grad : float
            limit value that avoids division by zero when computing the RMSProp preconditioner, denoted :math:`\eta` in the article
        history_weight : float
            weight of past values of :math:`v` in the exponential decay (cf RMSProp preconditioner), denoted :math:`\alpha` in the article
        selection_probas : np.ndarray of shape (2,)
            vector of selection probabilities for the MTM and PMALA kernels, respectively, i.e., :math:`[p_{MTM}, 1 - p_{MTM}]`
        k_mtm : int
            number of candidates in the MTM kernel, denoted :math:`K` in the article
        is_stochastic : bool
            if True, the algorithm performs sampling, and optimization otherwise, by default True
        compute_correction_term : bool
            wether or not to use the correction term (denoted :math:`\gamma` in the article) during the sampling (only used if `is_stochastic=True`), by default True
        """
        assert initial_step_size > 0
        assert extreme_grad > 0
        assert 0 < history_weight < 1

        assert isinstance(k_mtm, int) and k_mtm >= 1

        if isinstance(selection_probas, list):
            selection_probas = np.array(selection_probas)

        assert np.all(0 <= selection_probas)
        assert selection_probas.sum() == 1

        assert isinstance(is_stochastic, bool)
        assert isinstance(compute_correction_term, bool)

        self.initial_step_size = initial_step_size
        self.extreme_grad = extreme_grad
        self.history_weight = history_weight

        self.selection_probas = selection_probas
        self.k_mtm = k_mtm

        self.is_stochastic = is_stochastic
        self.compute_correction_term = compute_correction_term



import abc

import numpy as np


class Scaler(abc.ABC):
    r"""class that defines a transition between the sampling space and a more user friendly / interpretable space"""

    @abc.abstractmethod
    def from_scaled_to_lin(self, Theta_scaled: np.ndarray) -> np.ndarray:
        r"""applies the transformation from sampling scale to user friendly scale

        Parameters
        ----------
        Theta_scaled : np.ndarray of shape (-1, D)
            array in sampling scale

        Returns
        -------
        np.ndarray of shape (-1, D)
            array in user friendly scale
        """
        pass

    @abc.abstractmethod
    def from_lin_to_scaled(self, Theta_linscale: np.ndarray) -> np.ndarray:
        r"""applies the transformation from user friendly scale to sampling scale

        Parameters
        ----------
        Theta_linscale : np.ndarray of shape (-1, D)
            array in user friendly scale

        Returns
        -------
        np.ndarray of shape (-1, D)
            array in sampling scale
        """
        pass
